#
#  utils/stdio.py
#  bxtorch
#
#  Created by Oliver Borchert on May 11, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
import sys
import time
import datetime
import numpy as np

class ProgressBar:
    """
    The progress bar can be used to print progress for a specific task where
    N work items have to be processed. The progress bar also measures the
    execution time and provides an estimated remaining time for the  operation.

    A common use case for the progress bar are for-loops where one work item is
    completed per iteration. The progress bar is intended to be used either 
    from within a with statement or a for-loop.
    """

    # MARK: Static Methods
    @staticmethod
    def frac(num, denom):
        """
        Initializes a new progress bar with ceil(num/denom) work items.

        Parameters:
        -----------
        - num: int
            Numerator of the fraction.
        - denom: int
            Denominator of the fraction.

        Returns:
        --------
        - bxtorch.utils.ProgressBar
            The progress bar.
        """
        return ProgressBar(int(np.ceil(num / denom)))

    # MARK: Initialization
    def __init__(self, total, file=None):
        """
        Initializes a new progress bar with the given number of work items.

        Parameters:
        -----------
        - total: int
            The number of work items to be processed.
        - file: str, default: None
            If given, defines the file where the progress bar should write to
            instead of the command line. Intermediate directories are created
            automatically.
        """
        self.current = 0
        self.total = total
        self.tic = None
        self._counter = None
        if file is not None:
            os.makedirs(file, exist_ok=True)
            self.stream = open(file, 'a+')
        else:
            self.stream = sys.stdout

    # MARK: Instance Methods
    def start(self):
        """
        Starts to record the progress of the operation. Time measuring is
        initiated and the beginning of the operation is indicated on the
        command line.

        This method should never be called explicitly. It is implicitly called
        at the beginning of a with statement.
        """
        self.tic = time.time()
        self._print_progress(compute_eta=False)

    def step(self):
        """
        Tells the progress bar that one work item has been processed. The 
        command line output is updated as well as the estimated finishing time
        of the operation.

        If used from within a with statement, this method must be called
        explicitly, otherwise, it should not be called.
        """
        self.current += 1
        self._print_progress()

    def finish(self, metrics={}):
        """
        Stops the progress bar and prints the total duration of the operation.
        If metrics are given, these will be printed along with the duration.

        If metrics are given, this method must be called explicitly, otherwise,
        it is implicitly called at the end of a with statement or for loop.
        
        Note:
        -----
        Do not call this method multiple times.

        Parameters:
        -----------
        - metrics: dict, default: {}
            The metrics to print as key-value pairs. Usually, they provide more
            information about the operation whose progress has been tracked.
        """
        toc = time.time()
        self._print_done(metrics)
        self.tic = None

    # MARK: Special Methods
    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.tic is not None:
            self.finish()
        return False

    def __iter__(self):
        return self

    def __next__(self):
        if self._counter is None:
            self.start()
            self._counter = 0
            return self._counter
        elif self._counter < self.total - 1:
            self.step()
            self._counter += 1
            return self._counter
        else:
            self.finish()
            raise StopIteration

    # MARK: Private Methods
    def _print_progress(self, compute_eta=True):
        perc = self.current / self.total
        p = int(np.round(perc * 30))
        progress = "" if p == 0 else "=" * (p - 1) + ">" if p < 30 else "=" * 30
        whitespace = " " * (30 - p)
        elapsed = time.time() - self.tic
        if compute_eta:
            eta = datetime.timedelta(0, int((1 - perc) / perc * elapsed))
        else:
            eta = 'n/a'
        print(
            " [{}{}] ({:02.1%}) ETA {}".format(progress, whitespace, perc, eta),
            end='\r',
            file=self.stream
        )
        self.stream.flush()

    def _print_done(self, metrics):
        elapsed = datetime.timedelta(0, int(time.time() - self.tic))
        m_strings = []
        for k, v in sorted(metrics.items(), key=lambda k:k[0]):
            split = k.split('__')
            if len(split) > 1:
                f = split[1]
            else:
                f = '{:.5f}'
            string = f'{split[0]}: {f}'.format(v)
            m_strings += [string]
        print_text = " [Elapsed {}] {}".format(elapsed, ", ".join(m_strings))
        if len(print_text) < 54:
            print_text += " " * (54 - len(print_text))
        print(print_text, file=self.stream)
        self.stream.flush()

    # MARK: Special Methods
    def __del__(self):
        if self.stream != sys.stdout:
            self.stream.close()


def ensure_valid_directories(file):
    """
    Creates intermediate directories to ensure that files can be saved to their
    specified directory.

    Parameters:
    -----------
    - file: str
        The absolute path to the file whose directory structure to ensure.
    """
    dir = "/".join(file.split('/')[:-1])
    os.makedirs(dir, exist_ok=True)



#
#  graph/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .build import compute_transition_counts, assemble_adjacency_matrix
from .backend import Graph, TensorGraph
from .walker import BiasedSecondOrderRandomWalker



#
#  nn/utils/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .model import Config, Configurable
from .predictor import Predictor



#
#  graph/build.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import numpy as np
import scipy.sparse as sp

def compute_transition_counts(random_walks, num_nodes,
                              ensure_no_loops=False):
    """
    Computes the transition counts between all pairs of nodes based on the
    provided random walks.

    Parameters:
    -----------
    - random_walks: numpy.ndarray [W, L] or [W, L, N]
        A set of random walks where W is the number of walks and L is the walks'
        lengths. Random walks must either be given as integers representing 
        node indices or one-hot vectors describing the node.
    - num_nodes: int
        The number of nodes in the resulting graph. Must be equal to the third
        dimension of the provided random walks if present.
    - ensure_no_loops: bool, default: False
        Whether to assert that the transition matrix's diagonal is all zero.
        If the matrix's diagonal contains values, the random walker did not move
        in every step.

    Returns:
    --------
    - scipy.sparse.csr_matrix [N, N]
        The symmetric matrix containing the transition counts for all node
        pairs.
    """
    assert len(random_walks.shape) == 2 or random_walks.shape[-1] == num_nodes,\
        "The dimensionality of the random walks does not match the number " + \
        "of expected nodes."

    # 1) Prepare walks
    # 1.1) Make sure random walks provide node indices
    if len(random_walks.shape) == 3:
        random_walks = np.argmax(random_walks, axis=2)

    # 1.2) Get transitions
    random_walk_transitions = np.array([
        list(zip(random_walks[:,:-1], random_walks[:,1:]))
    ])[0].transpose(0, 2, 1).reshape(-1, 2)

    # 1.3) Ensure symmetric transitions
    random_walk_transitions = np.vstack([
        random_walk_transitions,
        np.roll(random_walk_transitions, shift=1, axis=1)
    ])

    # 2) Compute transition matrix
    transition_matrix = sp.csr_matrix(
        (np.ones(len(random_walk_transitions)), 
         (random_walk_transitions[:,0], random_walk_transitions[:,1])),
        shape=(num_nodes, num_nodes)
    )

    if ensure_no_loops:
        assert (np.diag(transition_matrix) == 0).all(), \
            "Diagonal of transition matrix is not all zero."

    return transition_matrix.tocsr()


def assemble_adjacency_matrix(transition_counts, num_edges,
                              inplace=True, seed=None):
    """
    Computes an adjacency matrix for a graph based on the given transition
    counts and the desired number of edges. The resulting adjacency matrix will
    represent a graph with no singleton nodes (however, possible with
    multiple connected components).

    Note:
    -----
    The strategy is described in *NetGAN: Generating Graphs via Random Walks* 
    (Bojchevski, Shchur, Zügner, Günnemann -- 2018).

    Parameters:
    -----------
    - transition_counts: scipy.sparse.csr_matrix [N, N]
        The transition counts (e.g. obtained from random walks) for all pairs
        of nodes. Must be symmetric.
    - num_edges: int
        The number of edges the output adjacency matrix should contain.
    - inplace: bool, default: True
        Whether the transition_counts matrix may be modified. Otherwise, a copy
        is performed.
    - seed: int, default: None
        The seed to use for generating random values.

    Returns:
    --------
    - scipy.sparse.csr_matrix
        A binary adjacency matrix containing the desired number of edges. The 
        function tries to assemble a matrix with ``2 * num_edges`` entries.
        However, if ``num_edges < transitions_count.shape[0]``, then this cannot
        be guaranteed.  The diagonal of the adjacency matrix is always zero.
    """
    # 1) Setup
    randomizer = np.random.RandomState(seed)

    # 1.1) Copy if needed
    if not inplace:
        transition_counts = transition_counts.copy()

    # 1.2) Set diagonal to zero
    transition_counts = transition_counts.tolil()
    transition_counts.setdiag(0)

    # 2) Check if the transition matrix can be converted easily
    if len(transition_counts.nonzero()[0]) // 2 <= num_edges:
        transition_counts[transition_counts.nonzero()] = 1
        return transition_counts

    # 3) Assemble the adjacency matrix according to paper
    N = transition_counts.shape[0]
    result = sp.dok_matrix((N, N))
    # transition probabilities
    P = (transition_counts / transition_counts.sum(axis=0)).T
    P[np.isnan(P)] = 0

    # 3.1) Iterate over nodes in random order to sample one neighbor
    for node in randomizer.permutation(N)[:min(num_edges, N)]:
        # 3.1.1) Skip if no neighbor for the node is present
        if np.sum(P[node]) == 0:
            continue

        # 3.1.2) Sample neighbor according to probabilities
        neighbor = randomizer.choice(N, p=P[node].A1)
        result[node, neighbor] = result[neighbor, node] = 1

    # 3.2) Sample remaining edges
    # 3.2.1) Compute probabilities for drawing
    num_remaining_edges = int(num_edges - result.sum() / 2)
    if num_remaining_edges > 0:
        # equals size of the upper triangular matrix
        num_choices = (N * N + N) // 2
        transition_counts[result.nonzero()] = 0
        P_triu = sp.triu(transition_counts).tocsr()
        P_triu_indices = np.triu_indices_from(transition_counts)
        probabilities = (P_triu / P_triu.sum())[P_triu_indices]

        # 3.2.2) Choose edges
        edges = randomizer.choice(
            num_choices, replace=False, p=probabilities.A1, size=num_remaining_edges
        )

        # 3.2.3) Add edge choices to result
        rows = P_triu_indices[0][edges]
        cols = P_triu_indices[1][edges]
        result[rows, cols] = result[cols, rows] = 1

    return result.tocsr()
    


#
#  nn/callbacks/schedule.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch
from .base import TrainingCallback

class ParameterScheduler(TrainingCallback):
    """
    The parameter scheduler is able to change the value of a variable
    over the course of the training. The scheduler modifies parameters
    **in-place**, hence, you must pass tensors to be modified and you must
    never pass them to the CPU.
    """

    # MARK: Initialization
    def __init__(self, parameter, schedule):
        """
        Initalizes a new scheduler for the given parameter.

        Parameters:
        -----------
        - parameter: torch.Tensor
            The parameter which should be modified over the course of the
            training.
        - schedule: func (float, int) -> float
            Function which should update the parameter (given as first
            argument) based on itself and the current epoch (second argument).
            The function must return the updated parameter. The scheduler
            function is called after every epoch.
        """
        self.parameter = parameter
        self.schedule = schedule
        self.epoch = None

    # MARK: Instance Methods
    def before_epoch(self, current, num_iterations):
        self.epoch = current

    def after_epoch(self, metrics):
        update = self.schedule(self.parameter.item(), self.epoch)
        self.parameter.set_(torch.tensor(update))

    def after_training(self):
        self.epoch = None



#
#  nn/functional/metrics.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch
import sklearn.metrics as metrics

def accuracy(y_pred, y_true):
    """
    Computes the accuracy of the class predictions.

    Parameters:
    -----------
    - y_pred: torch.LongTensor [N] or torch.FloatTensor [N, C]
        The class predictions made by the model. Can be either specific classes
        or predictions for each class.
    - y_true: torch.LongTensor [N]
        The actual classes.

    Returns:
    --------
    - torch.FloatTensor
        The accuracy.
    """
    y_pred = _ensure_classes(y_pred)
    return (y_pred == y_true).float().mean()


def recall(y_pred, y_true, c=1):
    """
    Computes the recall score of the class predictions.

    Parameters:
    -----------
    - y_pred: torch.LongTensor [N] or torch.FloatTensor [N, C]
        The class predictions made by the model. Can be either specific classes
        or predictions for each class.
    - y_true: torch.LongTensor [N]
        The actual classes.
    - c: int, default: 1
        The class to calculate the recall score for. Default assumes a binary
        classification setting.

    Returns:
    --------
    - torch.FloatTensor
        The recall score.
    """
    y_pred = _ensure_classes(y_pred)

    y_pred = y_pred == c
    y_true = y_true == c

    correct = (y_pred[y_true == y_pred]).sum()
    true_correct = y_true.sum()
    
    return correct.float() / true_correct.float()


def precision(y_pred, y_true, c=1):
    """
    Computes the precision score of the class predictions.

    Parameters:
    -----------
    - y_pred: torch.LongTensor [N] or torch.FloatTensor [N, C]
        The class predictions made by the model. Can be either specific classes
        or predictions for each class.
    - y_true: torch.LongTensor [N]
        The actual classes.
    - c: int, default: 1
        The class to calculate the recall score for. Default assumes a binary
        classification setting.

    Returns:
    --------
    - torch.FloatTensor
        The precision score.
    """
    y_pred = _ensure_classes(y_pred)

    y_pred = y_pred == c
    y_true = y_true == c

    correct = (y_pred[y_true == y_pred]).sum()
    true_correct = y_pred.sum()
    
    return correct.float() / true_correct.float()


def f1_score(y_pred, y_true, c=1):
    """
    Computes the F1-score of the class predictions.

    Parameters:
    -----------
    - y_pred: torch.LongTensor [N] or torch.FloatTensor [N, C]
        The class predictions made by the model. Can be either specific classes
        or predictions for each class.
    - y_true: torch.LongTensor [N]
        The actual classes.
    - c: int, default: 1
        The class to calculate the recall score for. Default assumes a binary
        classification setting.

    Returns:
    --------
    - torch.FloatTensor
        The F1-score.
    """
    y_pred = _ensure_classes(y_pred)
    p = precision(y_pred, y_true, c=c)
    r = recall(y_pred, y_true, c=c)
    return (2 * p * r) / (p + r)


def roc_auc_score(y_pred, y_true, c=1):
    """
    Computes the area under the ROC curve.

    Parameters:
    -----------
    - y_pred: torch.FloatTensor [N]
        The (binary) predictions made by the model.
    - y_true: torch.LongTensor [N]
        The actual classes.

    Returns:
    --------
    - torch.FloatTensor
        The ROC-AUC score.
    """
    assert y_pred.dim() == 1, \
        "ROC-AUC score only works in the binary case."

    return torch.as_tensor(
        metrics.roc_auc_score(y_true.numpy(), y_pred.numpy())
    )


def average_precision(y_pred, y_true, c=1):
    """
    Computes the average precision of the model predictions.

    Parameters:
    -----------
    - y_pred: torch.FloatTensor [N]
        The (binary) predictions made by the model.
    - y_trye: torch.LongTensor [N]
        The actual classes.

    Returns:
    --------
    - torch.FloatTensor
        The average precision.
    """
    assert y_pred.dim() == 1, \
        "Average precision only work in the binary case."

    return torch.as_tensor(
        metrics.average_precision_score(y_true.numpy(), y_pred.numpy())
    )


def _ensure_classes(y):
    assert y.dim() <= 2, \
        f"Invalid dimensionality {y.dim()} of predictions."

    if y.dim() == 2:
        return torch.argmax(y)
    elif y.dtype == torch.float:
        return torch.round(y)
    return y



#
#  utils/torch.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
import torch

def gpu_device(gpu):
    """
    Returns a device based on the passed parameters.
    
    Parameters:
    -----------
    - gpu: bool or int
        If int, the returned device is the GPU with the specified ID.
        If False, the returned device is the CPU, if True, the returned
        device is given as the GPU with the highest amount of free memory.
        
    Returns:
    --------
    - torch.device
        A PyTorch device.
    """
    if isinstance(gpu, bool) and gpu:
        assert torch.cuda.is_available()
        gpu = np.argmin([
            torch.cuda.memory_cached(torch.device(f'cuda:{i}'))
            for i in range(torch.cuda.device_count())
        ])
        return torch.device(f'cuda:{gpu}')
    elif isinstance(gpu, bool):
        return torch.device('cpu')
    else:
        assert gpu < torch.cuda.device_count()
        return torch.device(f'cuda:{gpu}')


def to_device(device, *args):
    """
    Passes the given tensors to the specified device.

    Parameters:
    -----------
    - device: torch.device
        The device to pass tensors to.
    - args: varargs of (torch.Tensor or list of torch.Tensor)
        The tensors to pass to the specified device. Tensors may be given in 
        lists such that a single variable can easily be assigned to a list of
        tensors upon function return. The given tensors may also be None, then
        None is returned for that tensor.

    Returns:
    --------
    - list of (torch.Tensor or list of torch.Tensor)
        The given tensors passed to the specified device. If tensors were passed
        as lists, they are also returned as lists.
    """
    return _recursive_apply('to', lambda x: x.to(device), args)


def share_memory(*args):
    """
    Shares the memory of the given tensors.

    Parameters:
    -----------
    - args: varargs of (torch.Tensor or list of torch.Tensor)
        The tensors to share the memory.

    Returns:
    --------
    - list of (torch.Tensor or list of torch.Tensor)
        The tensors in shared memory.
    """
    return _recursive_apply('share_memory_', lambda x: x.share_memory_(), args)


def pin_memory(*args):
    """
    Pins the memory of the given tensors.

    Parameters:
    -----------
    - args: varargs of (torch.Tensor or list of torch.Tensor)
        The tensors to pin the memory.

    Returns:
    --------
    - list of (torch.Tensor or list of torch.Tensor)
        The tensors with pinned memory.
    """
    return _recursive_apply('pin_memory', lambda x: x.pin_memory(), args)


def _recursive_apply(attribute, function, *args):
    if len(args) > 1:
        return [_recursive_apply(attribute, function, t) for t in args]
    elif len(args) == 1 and isinstance(args[0], list):
        return [_recursive_apply(attribute, function, t) for t in args[0]]
    elif len(args) == 1 and isinstance(args[0], tuple):
        return tuple(
            [_recursive_apply(attribute, function, t) for t in args[0]]
        )
    elif len(args) == 1 and hasattr(args[0], attribute):
        return function(args[0])
    else:
        return None
    

def to_sparse_tensor(X):
    """
    Creates a coalesced sparse PyTorch tensor from the given matrix.

    Parameters:
    -----------
    - X: scipy.sparse.csr_matrix
        The matrix to obtain the PyTorch tensor from.

    Returns:
    --------
    - torch.sparse.FloatTensor
        The sparse tensor.
    """
    data = torch.from_numpy(X.data).float()
    X = torch.sparse.FloatTensor(
        torch.LongTensor(X.nonzero()), data, torch.Size(X.shape)
    )
    return X.coalesce()

_eyes = {}

def to_one_hot(X, n):
    """
    Creates a one-hot matrix from a set of indices.

    Parameters:
    -----------
    - X: torch.LongTensor [N, D]
        The indices to convert into one-hot vectors.
    - n: int
        The number of entries in the one-hot vectors.

    Returns:
    --------
    - torch.ByteTensor [N, D, n]
        The one-hot matrix.
    """
    if n not in _eyes:
        _eyes[n] = torch.eye(n)
    return _eyes[n][X]



#
#  graph/walker.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from abc import ABC, abstractmethod
import numpy as np
import torch
import bxtorch.utils.numpy as npu
import numba

class RandomWalker(ABC):
    """
    Base random walker class for different random walkers on graphs.
    """

    # MARK: Initialization
    def __init__(self, graph):
        """
        Initializes a new random walker on the specified graph.

        Parameters:
        -----------
        - graph: bxtorch.graph.Graph
            The graph to generate random walks for.
        """
        self.graph = graph

    # MARK: Instance Methods
    def sample_walks(self, num_walks, steps, **kwargs):
        """
        Function to sample a batch of walks.

        Parameters:
        -----------
        - num_walks: int
            The number of walks to generate.
        - steps: int
            The number of steps for the random walker. Must be positive.
        - kwargs: keyword arguments
            Arguments dependent on the subclass.

        Returns:
        --------
        - numpy.ndarray [num_walks, steps]
            The indices of the nodes sampled in the random walks.
        """
        assert steps > 0, \
            "Random walker must at least visit one node."
        assert num_walks > 0, \
            "Random walker must at least make one walk."

        result = np.zeros((num_walks, steps), dtype=np.int64)
        self._sample_walks(result, **kwargs)
        return result

    def walk_generator(self, steps, **kwargs):
        """
        Generator function which yields random walks infinitely often.

        Parameters:
        -----------
        - steps: int
            The number of steps for the random walker. Must be positive.
        - kwargs: keyword arguments
            Arguments dependent on the subclass.

        Returns:
        --------
        - numpy.ndarray [steps]
            The indices of the nodes sampled in the random walk.
        """
        assert steps > 0, \
            "Random walker must at least visit one node."

        while True:
            out = np.zeros(steps, dtype=np.int64)
            yield self._generate_random_walk(out, **kwargs)

    # MARK: Private Methods
    @abstractmethod
    def _generate_random_walk(self, out, **kwargs):
        pass

    def _sample_walks(self, out, **kwargs):
        for i in range(out.shape[0]):
            self._generate_random_walk(out[i], **kwargs)


class BiasedSecondOrderRandomWalker(RandomWalker):
    """
    Random walker sampling random walks by using the biased second-order
    strategy described in *node2vec: Scalable Feature Learning for Networks*
    (Grover and Leskovec -- 2016).

    The walker accepts the following keyword arguments:
    - p: float
        The *return parameter*. If it is > max(q, 1), then walks tend to not
        revisit visited nodes, if it is set to < min(q, 1), then the walks tend 
        to explore the initial node's neighborhood. Must be set to a positive 
        value.
    - q: float
        The *in-out parameter*. If is greater than 1, then walks tend to explore
        the local neighborhood, otherwise depth-first walking is more likely. 
        Must be set to a positive value.
    """

    # MARK: Initialization
    def __init__(self, graph):
        super().__init__(graph)
        self.cum_degrees = np.cumsum(graph.node_degrees)
        self.edges = np.array(graph.adjacency.nonzero()).T

    # MARK: Private Methods
    def _generate_random_walk(self, out, p, q):
        assert p > 0 and q > 0, \
            "Return parameter p and in-out parameter q must be greater 0."

        return _generate_biased_second_order_random_walk(
            self.cum_degrees, self.edges, self.graph.num_nodes, out, p, q
        )

    def _sample_walks(self, out, p, q):
        _generate_biased_second_order_batch(
            self.cum_degrees, self.edges, self.graph.num_nodes, out, p, q
        )
        
#################
### FUNCTIONS ###
#################
@numba.njit
def _generate_biased_second_order_random_walk(cum_degrees, edges, N, out, p, q):

    # 1) Choose random node as starting node
    out[0] = np.random.choice(N)

    # 2) Sample transitions
    prev_neighbors = None

    for i in range(1, out.shape[0]):
        node = out[i - 1]

        # 2.1) Find neighbors
        upper = cum_degrees[node]
        if node == 0:
            neighbors = edges[:upper, 1]
        else:
            lower = cum_degrees[node - 1]
            neighbors = edges[lower:upper, 1]

        # 2.2) If this is the first transition, choose random node from 
        #      neighbors
        if i == 1:
            idx = np.random.choice(len(neighbors))
            out[i] = neighbors[idx]
            prev_neighbors = neighbors
            continue

        # 2.3) Else, sample transition according to search bias alpha
        prev_node = out[i - 2]

        dist_0 = neighbors == prev_node
        dist_1 = npu.intersection_mask(neighbors, prev_neighbors)
        dist_2 = 1 - dist_0 - dist_1

        alpha = dist_0 / p + dist_1 + dist_2 / q
        alpha_bins = np.cumsum(alpha / alpha.sum())
        sample = np.random.rand()
        choice = np.sum(alpha_bins <= sample)
        out[i] = neighbors[choice]

        # 2.4) Prepare next transition
        prev_neighbors = neighbors

    return out

@numba.njit(parallel=True)
def _generate_biased_second_order_batch(cum_degrees, edges, N, out, p, q):
    for i in numba.prange(out.shape[0]):
        _generate_biased_second_order_random_walk(
            cum_degrees, edges, N, out[i], p, q
        )



#
#  graph/graph.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
import scipy.sparse as sp
import scipy.sparse.csgraph as gs
import bxtorch.graph as g
from bxtorch.utils.numpy import cantor_pairing
import numba

def nonexisting_edges(graph, seed=None):
    """
    Returns a generator iterating randomly over the non-existing edges of the 
    graph. All returned edges will be of the format that the index of the first 
    node is smaller than the index of the second node. No self-loops are 
    returned. Note that this generator expects that the user samples much fewer
    edges than the number of nonexisting edges. Otherwise, do not use this
    function.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to yield nonexisting edges from.
    - seed: int, default: None
        The seed to use for random sampling.
    
    Returns:
    --------
    - generator of tuple of (int, int)
        The edges which do not exist in the graph.
    """
    randomizer = np.random.RandomState(seed)

    sampled = set()
    count = 0
    total = (graph.num_nodes ** 2 - graph.num_nodes) // 2 - graph.num_edges

    while count < total:
        # Choose node
        n = randomizer.choice(graph.num_nodes)
        m = randomizer.choice(graph.num_nodes)

        # No self-loops and no edges that already exist
        if n == m or graph.adjacency[n, m] == 1:
            continue

        # Order node and check if already sampled
        edge = (min(n, m), max(n, m))
        digest = cantor_pairing(*edge)
        if digest in sampled:
            continue

        # Set already sampled
        sampled.add(digest)
        count += 1

        yield edge


def random_subgraph(graph, size, seed=None):
    """
    Computes a random subgraph of the given graph with the specified fraction of
    edges. The resulting subgraph is guaranteed to be connected.

    Note:
    -----
    Implementation modified from https://bit.ly/2wdQsrz.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to sample a subgrah from.
    - size: float
        The size of the graphs. The number of edges for the subgraph is given as
        ``size * graph.num_edges``.
    - seed: int, default: None
        The seed to use for random sampling.

    Returns:
    --------
    - bxtorch.graph.Graph
        The subgraph with number of edges proportional to ``size``.
    - numpy.ndarray [N, 2]
        The edges which are not in the subgraph but in the original graph.
        The edges are always given with the node having the smaller index first.
        Use ``graph.utils.symmetric_edge_list`` to obtain 'directed' edges.
    """
    assert size > 0 and size < 1, \
        "Invalid size. Must be in the range (0, 1)."
    assert gs.connected_components(graph.adjacency)[0] == 1, \
        "Given graph has more than one connected component."

    randomizer = np.random.RandomState(seed)

    # 0) Setup
    num_edges = int(graph.num_edges * size)

    # 1) Ensure that resulting graph is connected
    mst = gs.minimum_spanning_tree(graph.adjacency)
    mst = mst + mst.T
    mst[mst > 1] = 1
    ensure_edges = np.array(sp.triu(mst).nonzero()).T

    if ensure_edges.shape[0] > num_edges:
        raise ValueError(
            f"Size {size} is too small. Subgraph cannot be fully connected."
        )
    
    # 2) Sample from the remaining edges
    A_prime = graph.adjacency - mst
    remaining_edges = np.array(sp.triu(A_prime).nonzero()).T
    edge_perturbation = randomizer.permutation(remaining_edges.shape[0])
    edge_choices = edge_perturbation[:num_edges-ensure_edges.shape[0]]

    # 3) Build adjacency
    add_edges = remaining_edges[edge_choices]
    A = sp.csr_matrix(
        (np.ones(num_edges),
         (np.concatenate([ensure_edges[:,0], add_edges[:,0]]),
          np.concatenate([ensure_edges[:,1], add_edges[:,1]]))),
        shape=(graph.num_nodes, graph.num_nodes)
    )
    A = A + A.T
    A[A > 1] = 1
    subgraph = g.Graph(A, graph.features, graph.labels)
    
    # 4) Find all other edges
    A_diff = graph.adjacency - subgraph.adjacency
    excluded_edges = np.array(sp.triu(A_diff).nonzero()).T

    return subgraph, excluded_edges



#
#  nn/training/supervised.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import copy
from .trainer import BaseTrainer

class SupervisedTrainer(BaseTrainer):
    """
    Trainer to be used in a supervised learning setting.

    A dataset supplied to the ``train`` or ``evaluate`` method must supply 
    tuples with the following items:

    * x: input to the model
    * y: correct output

    For the ``predict`` method, the data must be supplied as follows:

    * x: input to the model

    Further, the ``train`` method requires the following parameters:

    - optimizer: torch.optim.Optimizer or str
        The optimizer to use for optimizing the model's weights. If a string is 
        given, the corresponding optimizer is used with default 
        hyperparameters. Be aware that this optimizer is kept in memory as long 
        as the trainer exists.
    - optimizer_init: func (torch.nn.Module) -> torch.optim.Optimizer or str
        If used for parallel training, this function must be supplied instead 
        of the optimizer if you do not want to use a default optimizer given by
        a string.
    - loss: func (torch.Tensor, torch.Tensor) -> float
        The loss function to use.
    """

    def __init__(self, model):
        super().__init__(model)
        self._grad_accum_cache = {}

    # MARK: Callback Overrides
    def before_epoch(self, current, num_iterations):
        self._grad_accum_cache['num_iterations'] = num_iterations
        self._grad_accum_cache['current_iteration'] = 0

    def after_batch(self):
        self._grad_accum_cache['current_iteration'] += 1

    def after_epoch(self, metrics):
        self._grad_accum_cache = {}
    
    # MARK: Private Methods
    def _train_batch(self, data, optimizer=None, loss=None,
                     gradient_accumulation_steps=1):
        if isinstance(optimizer, str):
            if optimizer not in self._cache:
                self._cache[optimizer] = self.optimizer(optimizer)
            optimizer = self._cache[optimizer]

        it = self._grad_accum_cache['current_iteration']
        if it == 0:
            optimizer.zero_grad()

        loss_func = loss

        x, y_true = data
        y_pred = self._forward(x)
        loss = loss_func(y_pred, y_true) / gradient_accumulation_steps

        loss.backward()

        it_is_last = it == self._grad_accum_cache['num_iterations'] - 1
        if it % gradient_accumulation_steps == 0 or it_is_last:
            optimizer.step()
            optimizer.zero_grad()

        return loss.item()

    def _predict_batch(self, data):
        x, y_true = data
        y_pred = self._forward(x)
        return y_pred, y_true

    def _forward(self, x):
        if isinstance(x, (list, tuple)):
            return self.model(*x)
        else:
            return self.model(x)



#
#  nn/callbacks/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .base import TrainingCallback, PredictionCallback, CallbackException
from .early_stopping import EarlyStopping
from .logging import EpochProgressLogger, BatchProgressLogger, \
    PredictionProgressLogger
from .parallel import SynchronizationCallback
from .saving import ModelSaverCallback
from .schedule import ParameterScheduler



#
#  job/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 15, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .job import Job
import bxtorch.job.arguments as args



#
#  nn/modules/rnn.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import torch
import torch.nn as nn

class StackedLSTM(nn.Module):
    """
    The stacked LSTM is an extension to PyTorch's native LSTM allowing stacked
    LSTMs with different hidden dimensions being stacked.
    """

    # MARK: Initialization
    def __init__(self, input_size, hidden_sizes, bias=True, batch_first=False):
        """
        Initializes a new stacked LSTM according to the given parameters.

        Parameters:
        -----------
        - input_size: int
            The dimension of the sequence's elements.
        - hidden_sizes: list of int
            The dimensions of the stacked LSTM's layers.
        - bias: bool, default: True
            Whether to use biases in the LSTM.
        - batch_first: bool, default: False
            Whether the batch or the sequence can be found in the first
            dimension.
        """
        super().__init__()
        self.batch_first = batch_first

        lstms = []
        dims = zip([input_size] + hidden_sizes, hidden_sizes)
        for i, (in_dim, out_dim) in enumerate(dims):
            lstms.append(nn.LSTM(
                in_dim, out_dim, bias=bias, batch_first=False
            ))
        self.lstms = nn.ModuleList(lstms)

    # MARK: Instance Methods
    def forward(self, inputs, initial_states=None, return_sequence=True):
        """
        Computes the forward pass through the stacked LSTM.

        Parameters:
        -----------
        - inputs: torch.FloatTensor [S, B, N]
            The inputs fed to the LSTM one after the other. Sequence length S,
            batch size B, and input size N. If ``batch_first`` is set to True,
            the first and second dimension should be swapped.
        - initial_states: list of tuple of 
                (torch.FloatTensor [H_i], torch.FloatTensor [H_i]),
                default: None
            The initial states for all LSTM layers. The length of the list must
            match the number of layers in the LSTM, the sizes of the states must
            match the hidden sizes of the LSTM layers. If None is given, the
            initial states are defaulted to all zeros.
        - return_sequence: bool, default: True
            Whether to return all outputs from the last LSTM layer or only the
            last one.

        Returns:
        --------
        - torch.FloatTensor [S, B, K] or torch.FloatTensor [B, K]
            Depending on whether sequences are returned, either all outputs
            or only the output from the last cell are returned. If the stacked
            LSTM was initialized with ``batch_first``, the first and second
            dimension are swapped when sequences are returned.
        """
        if self.batch_first:
            inputs = inputs.transpose(1, 0)

        sequence_length, batch_size = inputs.size()[0:2]

        if initial_states is not None:
            for i, state in enumerate(initial_states):
                initial_states[i] = (
                    state[0].view(1, batch_size, -1),
                    state[1].view(1, batch_size, -1)
                )
        states = initial_states or [None] * len(self.lstms)

        outputs = []
        for n, input_val in enumerate(inputs):
            output = None
            for i, lstm in enumerate(self.lstms):
                val = input_val if i == 0 else output
                val = val.view(1, batch_size, -1)
                output, states[i] = lstm(val, states[i])
            if return_sequence or n == sequence_length - 1:
                outputs.append(output)

        if return_sequence:
            result = torch.stack(outputs)
            if self.batch_first:
                result = result.transpose(1, 0)
            return result
        else:
            return outputs[0]



#
#  nn/utils/model.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import copy
from abc import ABC, abstractclassmethod
import json
import torch
import torch.nn as nn
from bxtorch.utils.stdio import ensure_valid_directories

class Config(ABC):
    """
    Base class which standardizes configuration for PyTorch modules.
    """

    # MARK: Static Methods
    @classmethod
    def load(cls, file):
        """
        Loads a configuration from the specified JSON file.

        Parameters:
        -----------
        - file: str
            The full path to the file. The extension does not need to be
            specified, but must be .json.

        Returns:
        --------
        - bxtorch.nn.Config
            The loaded configuration.
        """
        if not file.endswith('.json'):
            file += '.json'
        with open(file, 'r') as f:
            return cls(**json.load(f))

    @abstractclassmethod
    def _parameters(cls):
        """
        Returns the parameters and their optional values than can be specified
        for this configuration.

        Returns:
        --------
        - dict
            A mapping from parameter names to default values. If a parameter
            is required and has no default value, None should be specified.
        """
        pass
    
    # MARK: Initialization
    def __init__(self, **kwargs):
        """
        Initializes a new config from the given parameters.
        """
        all_params = type(self)._parameters()
        params = {}
        keys = set(all_params.keys())

        for k, v in kwargs.items():
            if not k in keys:
                raise ValueError(
                    f'Invalid configuration parameter {k}.'
                )
            params[k] = v
            keys.remove(k)

        for k in keys:
            if all_params[k] is None:
                raise ValueError(
                    f'Missing required configuration parameter {k}.'
                )
            params[k] = all_params[k]
        
        self._params = params

        if not self._is_valid():
            raise ValueError(
                f'Invalid configuration parameters. Check the documentation.'
            )

    # MARK: Instance Methods
    def save(self, file):
        """
        Saves the configuration to a JSON file.

        Parameters:
        -----------
        - file: str
            The full path to the file. The extension does not need to be
            specified, but must be .json.
        """
        ensure_valid_directories(file)
        if not file.endswith('.json'):
            file += '.json'
        with open(file, 'w+') as f:
            json.dump(self._params, f)

    # MARK: Private Methods
    def _is_valid(self):
        """
        Checks whether the given configuration is valid. This method should be
        overridden if required. The default implementation implies validity.

        Returns:
        --------
        - bool
            Whether the configuration is valid.
        """
        return True

    # MARK: Special Methods
    def __getstate__(self):
        return self._params

    def __setstate__(self, state):
        self._params = state

    def __repr__(self):
        return self._params.__repr__()

    def __getattr__(self, name):
        try:
            return self._params[name]
        except KeyError:
            class_name = self.__class__.__name__
            raise AttributeError(
                f'Configuration {class_name} has no parameter {name}.'
            )


class Configurable(ABC):
    """
    Mixin class to make torch.nn.Module configurable in an easy and standardized
    way. For initialization, a module is then given a single config file of 
    type ``bxtorch.nn.Config``. Properties of the configuration can be easily 
    accessed via ``self.<property>``. Additionally, this class makes it easy
    to save and load the module in a consistent way.

    When subclassing a PyTorch module, make sure to include the Configurable
    mixin as *first* dependency.

    Note:
    -----
    Only top-level modules should use this mixin to profit from saving/loading
    modules easily.
    """

    # MARK: Static Methods
    @classmethod
    def load(cls, file):
        """
        Loads the model by loading a configuration defining the architecture
        and another file containing the weights for this architecture.

        Parameters:
        -----------
        - file: str
            The full path to the file. The configuration will be loaded from
            <file>.json and the model's weights will be loaded from <file>.pt.

        Returns:
        --------
        - bxtorch.nn.Configurable
            The loaded model.
        """
        if cls._config_class() is not None:
            config = cls._config_class().load(file)
            model = cls(config)
        else:
            model = cls(None)
        params = torch.load(f'{file}.pt')
        model.load_state_dict(params)
        return model

    @abstractclassmethod
    def _config_class(cls):
        """
        Defines the config class which the module uses for configuration.

        Returns:
        --------
        - type like bxtorch.nn.Config
            The configuration class.
        """
        pass

    # MARK: Initialization
    def __init__(self, config):
        super(Configurable, self).__init__()
        self._config = config

    # MARK: Computed Properties
    @property
    def num_parameters(self):
        return sum(p.numel() for p in self.parameters())

    @property
    def num_trainable_parameters(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)

    # MARK: Instance Methods
    def save(self, file):
        """
        Saves the model and its configuration to two files.

        Parameters:
        -----------
        - file: str
            The full path to the file. The configuration will be saved to
            <file>.json while the model's weights will be saved to <file>.pt.
        """
        ensure_valid_directories(file)
        self._config.save(file)
        # Ensure saving CPU model
        state_dict = copy.deepcopy(self.state_dict())
        result = {}
        for k, v in state_dict.items():
            result[k] = v.cpu()
        torch.save(result, f'{file}.pt')

    # MARK: Special Methods
    def __getattr__(self, name):
        try:
            return super(Configurable, self).__getattr__(name)
        except AttributeError:
            return getattr(self._config, name)



#
#  nn/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 11, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .callbacks import *
from .data import *
from .utils import *
from .training import *
from .modules import *



#
#  graph/graph.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
import numpy.linalg as la
import scipy.sparse.csgraph as gs
import scipy.stats as st
import scipy.sparse as sp
from bxtorch.utils.numpy import intersection_mask_pairs

def edge_overlap(graph_lhs, graph_rhs):
    """
    Computes the edge overlap between two graphs.

    Parameters:
    -----------
    - graph_lhs: bxtorch.graph.Graph
        The first graph.
    - graph_rhs: bxtorch.graph.Graph
        The second graph.

    Returns:
    --------
    - int
        The number of overlapping edges.
    """

    A = graph_lhs.adjacency
    B = graph_rhs.adjacency

    A_nnz = np.array(sp.triu(A).nonzero()).T
    B_nnz = np.array(sp.triu(B).nonzero()).T

    return int(intersection_mask_pairs(A_nnz, B_nnz).sum())


def wedge_count(graph):
    """
    Computes the number of wedges (2-stars) in the graph.

    Note:
    -----
    The number of wedges s is given by the following formula:

        s = \sum_{u \in V}{\binom{d(u)}{2}} = 
            \sum_{u \in V}{1/2 d(u) * (d(u) - 1)}

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the wedge count for.

    Returns:
    --------
    - int
        The number of wedges in the graph.
    """
    degs = graph.node_degrees
    return np.sum(degs * (degs - 1) / 2)


def claw_count(graph):
    """
    Computes the number of claws (3-stars) in the graph.

    Note:
    -----
    The number of claws z is given by the following formula:

        z = \sum_{u \in V}{\binom{d(u)}{3}} = 
            \sum_{u \in V}{1/6 d(u) * (d(u) - 1) * (d(u) - 2)}

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the wedge count for.

    Returns:
    --------
    - int
        The number of wedges in the graph.
    """
    degs = graph.node_degrees
    return np.sum(degs * (degs - 1) * (degs - 2) / 6)


def triangle_count(graph):
    """
    Computes the number of triangles in the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the triangle count for.

    Returns:
    --------
    - int
        The number of triangles in the graph.
    """
    A3 = la.matrix_power(graph.adjacency, 3)
    return np.trace(A3) / 6


def square_count(graph):
    """
    Computes the number of squares in the graph.

    Note:
    -----
    Formula is taken from https://bit.ly/2YwBWr2.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the square count for.

    Returns:
    --------
    - int
        The square count.
    """
    A4 = la.matrix_power(graph.adjacency, 4)
    denom = np.sum(graph.node_degrees ** 2)
    return np.trace(A4) / denom


def edge_distribution_entropy(graph):
    """
    Computes the edge distribution entropy of the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the edge distribution entropy for.

    Returns:
    --------
    - float
        The edge distribution entropy in [0, 1].
    """
    degs = graph.node_degrees / (2 * graph.num_edges)
    return -1 / np.log(graph.num_nodes) * np.sum(degs * np.log(degs))


def gini_coefficient(graph):
    """
    Computes the Gini coefficient of the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the Gini coefficient for.

    Returns:
    --------
    - float
        The Gini coefficient in [0, 1].
    """
    degs = graph.node_degrees
    np.sort(degs)
    num = 2 * np.sum(np.arange(1, graph.num_nodes + 1) * degs)
    denom = graph.num_nodes * np.sum(degs)
    return num / denom - (graph.num_nodes + 1) / graph.num_nodes


def average_diameter(graph):
    """
    Computes the average diameter of the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the average diameter for.

    Returns:
    --------
    - float
        The average diameter.
    """
    shortest_paths = gs.shortest_path(graph.adjacency)
    mask = (~np.isinf(shortest_paths)) * (1 - np.eye(graph.num_nodes))
    return np.mean(shortest_paths[mask])


def power_law_exponent(graph):
    """
    Estimates the power law exponent of the graph.

    Note:
    -----
    The power law exponent gamma can be estimated as follows:

        gamma = 1 + \frac{n}{\sum_{u \in V}{\ln{\frac{d(u)}{d_{\min}}}}}

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to estimate the power law exponent for.

    Returns:
    --------
    - float
        The power law exponent.
    """
    degs = graph.node_degrees
    min_deg = max(np.min(degs), 1)
    denom = np.sum(np.log(degs / min_deg))
    return 1 + graph.num_nodes / denom


def clustering_coefficient(graph):
    """
    Computes the clustering coefficient of the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the clustering coefficient for.

    Returns:
    --------
    - float
        The clustering coefficient.
    """
    return 3 * triangle_count(graph) / claw_count(graph)


def degree_assortativity(graph):
    """
    Computes the degree assortativity of the graph.

    Parameters:
    -----------
    - graph: bxtorch.graph.Graph
        The graph to compute the clustering coefficient for.

    Returns:
    --------
    - float
        The degree assortativity in [-1, 1].
    """
    edges = graph.edges
    degs = graph.node_degrees
    return st.pearsonr(degs[edges[:,0]], degs[edges[:,1]])[0]



#
#  nn/callbacks/schedule.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
import copy
import torch
import bxtorch.nn as xnn
from .base import TrainingCallback
from bxtorch.utils.stdio import ensure_valid_directories

class ModelSaverCallback(TrainingCallback):
    """
    The callback stores the trained model after every epoch with a unique
    name per epoch. If the model uses the ``bxtorch.nn.Configurable`` mixin,
    its config and state dict are stored after every epoch, otherwise only
    its state dict.
    """

    # MARK: Initialization
    def __init__(self, directory, file_template='model_epoch-{}'):
        """
        Initializes a new ModelSaverCallback.

        Parameters:
        -----------
        - directory: str
            The directory where the models should be stored.
        - file_template: str, default: 'model_epoch_{}'
            A file template that can be formatted with a single integer, i.e.
            the epoch.
        """
        self.file_template = f'{directory}/{file_template}'
        self.model = None
        self.epoch = None

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        self.model = model

    def before_epoch(self, current, num_iterations):
        self.epoch = current

    def after_epoch(self, metrics):
        file = self.file_template.format(self.epoch)
        if isinstance(self.model, xnn.Configurable):
            self.model.save(file)
        else:
            ensure_valid_directories(file)
            state_dict = copy.deepcopy(self.model.state_dict())
            result = {}
            for k, v in state_dict.items():
                result[k] = v.cpu()
            torch.save(result, f'{file}.pt')
        self.epoch = None

    def after_training(self):
        self.model = None



#
#  graph/graph.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from abc import ABC, abstractmethod
import functools
import numpy as np
import scipy.sparse as sp
import scipy.sparse.csgraph as gs
import torch
import bxtorch.utils.torch as ut
import bxtorch.graph.stats as stats
import bxtorch.graph.func as func

class BaseGraph(ABC):
    """
    Abstract base class for graph implementations using matrices.
    """

    # MARK: Initialization
    def __init__(self, adjacency, features, labels):
        """
        Initializes a new graph. The types of the passed inputs depend on the
        particular implementation of the subclass.

        Parameters:
        -----------
        - adjacency: matrix [N, N]
            The graph's symmetric adjacency matrix. Unless the graph is small or
            dense, it should be sparse.
        - features: matrix [N, D]
            The graph's feature matrix. Should only be sparse if the features 
            are sparse. The parameter may be None if the graph does not have
            node features.
        - labels: vector [N]
            The labels for all nodes. May be None if the graph does not have
            node labels.
        """
        assert adjacency.shape[0] == adjacency.shape[1], \
            "Adjacency matrix is not square."
        assert features is None or adjacency.shape[0] == features.shape[0], \
            "Dimensions of adjacency matrix and feature matrix are invalid."
        assert labels is None or adjacency.shape[0] == labels.shape[0], \
            "Dimension of label matrix is invalid."
        
        self.adjacency = adjacency
        self.features = features
        self.labels = labels

    # MARK: Computed Properties
    @property
    def num_nodes(self):
        """
        Computes the number of nodes in the graph.

        Returns:
        --------
        - int
            The number of nodes.
        """
        return self.adjacency.shape[0]

    @property
    @abstractmethod
    def num_edges(self):
        """
        Computes the number of edges in the graph.

        Returns:
        --------
        - int
            The number of edges.
        """
        pass

    @property
    @abstractmethod
    def num_classes(self):
        """
        Computes the number of classes assuming the node labels are present.

        Returns:
        --------
        - int
            The number of classes
        """
        pass

    # MARK: Special Methods
    def __repr__(self):
        name = self.__class__.__name__
        result = f'{name}(<nodes: {self.num_nodes}, edges: {self.num_edges}'
        if self.features is not None:
            result += f', features: {self.features.shape[1]}'
        if self.labels is not None:
            result += f', classes: {self.num_classes}'
        return result + '>)'

    def __len__(self):
        return self.adjacency.shape[0]


class Graph(BaseGraph):
    """
    The graph class represents undirected graphs, optionally with node features
    and labels. 

    The following datatypes are expected:
     * adjacency_matrix: numpy.ndarray or scipy.sparse.csr_matrix
     * feature_matrix: numpy.ndarray or scipy.sparse.csr_matrix
     * labels: numpy.ndarray
    """

    # MARK: Static Methods
    @staticmethod
    def load(file):
        """
        Loads a graph from the specified file.

        Note:
        -----
        Implementation adapted from https://github.com/danielzuegner/netgan.

        Parameters:
        -----------
        - file: str
            The file to load the graph from. The initialization depends on the
            extension of the file name.
        """
        if file.endswith('.npz'):
            with np.load(file, allow_pickle=True) as loader:
                loader = dict(loader)['arr_0'].item()
                A = sp.csr_matrix(
                    (loader['adj_data'], loader['adj_indices'],
                        loader['adj_indptr']),
                    shape=loader['adj_shape']
                )

                if 'attr_data' in loader:
                    X = sp.csr_matrix(
                        (loader['attr_data'], loader['attr_indices'],
                            loader['attr_indptr']),
                        shape=loader['attr_shape']
                    )
                else:
                    X = None

                Z = loader.get('labels')

                A = A + A.T
                A[A > 1] = 1
            
            return Graph(A, X, Z)
        raise ValueError(
            'Graph cannot be loaded if no filename extension is specified. '
            'Currently, valid extensions are [.npz].'
        )

    @staticmethod
    def _random(num_nodes, num_edges):
        A = sp.dok_matrix((num_nodes, num_nodes))
        i = 0
        while i < num_edges:
            n, m = (np.random.choice(num_nodes), np.random.choice(num_nodes))
            if n == m:
                continue
            n, m = (min(n, m), max(n, m))
            if A[n, m] == 1:
                continue
            A[n, m] = 1
            i += 1

        A = A.tocsr()
        A = A + A.T
        A[A > 1] = 1

        return Graph(A, None, None)

    # MARK: Computed Properties
    @property
    def num_edges(self):
        return int(self.adjacency.sum()) // 2

    @property
    def num_classes(self):
        return np.max(self.labels) + 1

    @property
    def node_degrees(self):
        return self.adjacency.sum(axis=1).A1.astype(np.int64)

    @property
    def edges(self):
        return np.array(self.adjacency.nonzero()).T

    # MARK: Instance Methods
    def largest_connected_component(self):
        """
        Returns a new graph containing only the nodes from the largest connected
        component.

        Returns:
        --------
        - bxtorch.graph.Graph
            A new, potentially smaller, graph.
        """
        _, components = gs.connected_components(self.adjacency)
        largest_component = np.bincount(components).argmax()
        mask = components == largest_component
        return Graph(
            self.adjacency[mask][:, mask], 
            self.features[mask] if self.features is not None else None, 
            self.labels[mask] if self.labels is not None else None
        )

    def save(self, file):
        """
        Saves the graph to the specified file. The file should have no
        extension such that it can be saved with the .npz extension.

        Parameters:
        -----------
        - file: str
            The file to save the graph to.
        """
        data = {
            'adj_data': self.adjacency.data,
            'adj_indices': self.adjacency.indices,
            'adj_indptr': self.adjacency.indptr,
            'adj_shape': self.adjacency.shape,
            'attr_data': self.features.data,
            'attr_indices': self.features.indices,
            'attr_indptr': self.features.indptr,
            'attr_shape': self.features.shape,
            'labels': self.labels
        }
        np.savez(file, data)

    # MARK: Special Methods
    def __getattr__(self, name):
        # Check if there's a stats function or a func function
        if hasattr(stats, name):
            f = getattr(stats, name)
        elif hasattr(func, name):
            f = getattr(func, name)
        else:
            classname = self.__class__.__name__
            raise AttributeError(
                f'Cannot get attribute {name} of {classname}.'
            )
        return functools.partial(f, self)

    def __eq__(self, other):
        if not isinstance(other, Graph):
            return False
        if other.num_nodes != self.num_nodes:
            return False
        return not (self.adjacency != other.adjacency).todense().any()


class TensorGraph(BaseGraph):
    """
    The graph class represents undirected graphs, optionally with node features
    and labels as PyTorch (sparse) tensors. Some operations can be implemented more efficiently with this graph. Further, this implementation can be used
    more easily with PyTorch models.

    The following datatypes are expected:
     * adjacency_matrix: torch.FloatTensor or torch.sparse.FloatTensor
     * feature_matrix: torch.FloatTensor or torch.sparse.FloatTensor
     * labels: torch.FloatTensor
    """

    # MARK: Static Methods
    @staticmethod
    def from_graph(graph):
        """
        Initializes a new PyTorch graph given a simple graph.

        Parameters:
        -----------
        - graph: bxtorch.graph.Graph
            The graph to initialize this PyTorch graph from.

        Returns:
        --------
        - bxtorch.graph.TensorGraph
            A newly initialized PyTorch graph.
        """
        A = graph.adjacency
        if isinstance(A, sp.csr_matrix):
            A = ut.to_sparse_tensor(A)
        else:
            A = torch.from_numpy(A).float()

        X = graph.features
        if isinstance(X, sp.csr_matrix):
            X = ut.to_sparse_tensor(X)
        elif X is not None:
            X = torch.from_numpy(X).float()

        Z = graph.labels
        if Z is not None:
            Z = torch.from_numpy(Z).float()
        
        return TensorGraph(A, X, Z)

    # MARK: Computed Properties
    @property
    def num_edges(self):
        return len(self.adjacency.values()) // 2
    
    @property
    def num_classes(self):
        return int(torch.max(self.labels) + 1)

    @property
    def node_degrees(self):
        return torch.sparse.sum(self.adjacency, dim=1).values().long()

    @property
    def edges(self):
        return self.adjacency.indices().transpose(0, 1)
    
    # MARK: Instance Methods
    def to(self, device):
        """
        Moves the graph to the specified device and returns it.

        Parameters:
        -----------
        - device: torch.device
            The device the graph should be moved to.

        Returns:
        --------
        - bxtorch.graph.TensorGraph
            A new graph moved to specified device.
        """
        return TensorGraph(
            *ut.to_device(self.adjacency, self.features, self.labels)
        )



#
#  nn/data/infinite_loader.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
import sys
import torch
import torch.multiprocessing as mp
from bxtorch.utils.torch import share_memory, pin_memory

def _default_collate_fn(items):
    return torch.stack(items)

class InfiniteDataLoader:
    """
    The infinite data loader works analogously to PyTorch's DataLoader. However,
    instead of being provided with a fixed-size dataset, it uses an infinite
    dataset from which infinitely many items can be sampled.
    """

    # MARK: Initialization
    def __init__(self, dataset, batch_size=1, num_workers=0, prefetch=3,
                 pin_memory=False, collate_fn=_default_collate_fn):
        """
        Initializes a new infinite data loader.

        Parameters:
        -----------
        - dataset: bxtorch.nn.InfiniteDataset
            The infinite dataset to sample from.
        - batch_size: int
            The number of items to use in every batch.
        - num_workers: int
            The number of workers to use for loading data. If set to 0, items
            are sampled from the dataset on the main thread. Otherwise, as many
            workers are used on background threads to sample from the dataset
            collaboratively. If set to None, the number of workers equals the
            number of cores on the computer.
        - prefetch: int, default: 3
            A number indicating how many batches should be precomputed at any
            given time.
        - pin_memory: bool, default: False
            Whether to pin data in CUDA pinned memory.
        - collate_fn: func (objects) -> torch.Tensor
            The function used to aggregate samples obtained from a dataset to
            a batch.
        """
        assert prefetch >= 1, "The prefetch must be at least 1."

        self.dataset = dataset
        self.iterator = iter(dataset)
        self.batch_size = batch_size
        self.pin_memory = pin_memory
        self.collate_fn = collate_fn

        if num_workers is None:
            self.num_workers = os.cpu_count()
        else:
            self.num_workers = num_workers
        
        if self.num_workers != 0:
            self.distribution_queue = mp.Queue()
            self.worker_queue = mp.Queue()
            self.result_queue = mp.Queue()
            self.batch_queue = mp.Queue()

            # 1) Initialize distributer
            self.distributer = mp.Process(
                target=_distribution_function,
                args=(self.num_workers, self.distribution_queue,
                      self.worker_queue)
            )
            self.distributer.daemon = True
            self.distributer.start()

            # 2) Initialize workers
            self.workers = []
            for _ in range(self.num_workers):
                worker = mp.Process(
                    target=_worker_function,
                    args=(self.iterator, self.worker_queue, self.result_queue,
                          self.pin_memory)
                )
                worker.daemon = True
                worker.start()
                self.workers.append(worker)

            # 3) Initialize aggregator
            self.aggregator = mp.Process(
                target=_aggregator_function,
                args=(self.batch_size, self.result_queue, self.batch_queue,
                      self.collate_fn)
            )
            self.aggregator.daemon = True
            self.aggregator.start()

            # 4) Prefetch data
            self.prefetch = prefetch
            self.expected_batch_count = 0
            self._prefetch_batch(
                decrement=False
            )

    # MARK: Private Methods
    def _prefetch_batch(self, decrement=True):
        if decrement:
            self.expected_batch_count -= 1
        if self.expected_batch_count < self.prefetch:
            num = self.batch_size * self.prefetch
            self.expected_batch_count += self.prefetch
        else:
            return
        self.distribution_queue.cancel_join_thread()
        self.distribution_queue.put(num)

    # MARK: Special Methods
    def __iter__(self):
        return self

    def __next__(self):
        if self.num_workers == 0:
            items = [next(self.iterator) for _ in range(self.batch_size)]
            return self.collate_fn(items)
        else:
            self._prefetch_batch()
            return self.batch_queue.get()

    def __del__(self):
        if hasattr(self, 'workers'):
            # 1) Shutdown distributer
            self.distribution_queue.put(None)
            self.distribution_queue.close()

            # 2) Shutdown workers
            for _ in range(self.num_workers):
                self.worker_queue.put(None)
            self.worker_queue.close()

            # 3) Shutdown aggregator
            self.result_queue.put(None)
            self.result_queue.close()
            self.batch_queue.close()

            # 4) Ensure graceful shutdown
            for worker in self.workers:
                worker.join()
            self.aggregator.join()


def _distribution_function(num_workers, distribution_queue, worker_queue):
    while True:
        num = distribution_queue.get()
        if num is None:
            return
        length = num // num_workers
        for i in range(num_workers):
            if i == num_workers - 1:
                count = num - length * (num_workers - 1)
            else:
                count = length
            worker_queue.cancel_join_thread()
            worker_queue.put(count)

def _worker_function(generator, worker_queue, result_queue, pin_memory):
    while True:
        num_samples = worker_queue.get()
        if num_samples is None:
            return

        items = [next(generator) for _ in range(num_samples)]
        if pin_memory:
            items = pin_memory(items)
        else:
            items = share_memory(items)

        result_queue.cancel_join_thread()
        result_queue.put(items)
    
def _aggregator_function(batch_size, result_queue, batch_queue, collate_fn):
    int_result = None
    count = 0
    while True:

        while count < batch_size:
            items = result_queue.get()

            if items is None:
                return
            
            if int_result is None:
                int_result = items
            else:
                int_result += items

            count += len(items)

        batch = int_result[:batch_size]
        int_result = int_result[batch_size:]
        count -= batch_size

        batch_queue.cancel_join_thread()
        batch_queue.put(collate_fn(batch))



#
#  nn/data/noise.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .infinite_dataset import InfiniteDataset
import bxtorch.nn.functional as X

class NoiseDataset(InfiniteDataset):
    """
    Infinite dataset for generating noise from a given probability distribution.
    Should be used with Generative Adversarial Networks.
    """

    # MARK: Initialization
    def __init__(self, noise_type, dimension):
        """
        Initializes a new dataset with the given noise type.

        Parameters:
        -----------
        - noise_type: str
            The noise type to use.
        - dimension: int
            The dimension of the noise to generate.
        """
        super().__init__(self._generator_func(noise_type, dimension))

    # MARK: Generator
    def _generator_func(self, noise_type, dimension):
        while True:
            yield X.generate_noise([dimension], noise_type)



#
#  utils/numpy.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
import numba

@numba.njit
def intersection_mask(x, y):
    """
    Computes the intersection of the given arrays.

    The arrays are expected to be sorted in ascending order and elements must
    be unique. Otherwise, the result is undefined.

    Parameters:
    -----------
    - x: numpy.ndarray [N]
        Array number one.
    - y: numpy.ndarray [M]
        Array number two.

    Returns:
    --------
    - numpy.ndarray [N]
        A mask for the array passed as first parameter. When applied to the
        first array, it returns the values contained in both arrays.
    """
    mask = np.zeros_like(x)
    
    N, M = (x.shape[0], y.shape[0])
    xp, yp = (0, 0)

    while xp < N and yp < M:
        if x[xp] == y[yp]:
            mask[xp] = 1
            xp += 1
            yp += 1
        elif x[xp] < y[yp]:
            xp += 1
        else:
            yp += 1

    return mask


@numba.njit
def intersection_mask_pairs(x, y):
    """
    Computes the intersection of the given pair arrays.

    The arrays are expected to be sorted alphabetically (i.e. (1, 0) comes 
    after (0, 9)) in ascending order (first in terms of the first dimension,
    then in terms of the second). Otherwise, the result is undefined.

    Parameters:
    -----------
    - x: numpy.ndarray [N, 2]
        Array of pairs number one.
    - y: numpy.ndarray [N, 2]
        Array of pairs number two.

    Returns:
    --------
    - numpy.ndarray [N]
        A mask for the array passed as first parameter. When applied to the
        first array, it returns the pairs contained in both arrays.
    """
    N, M = (x.shape[0], y.shape[0])
    mask = np.zeros(N)

    xp, yp = (0, 0)

    while xp < N and yp < M:
        if x[xp][0] == y[yp][0] and x[xp][1] == y[yp][1]:
            mask[xp] = 1
            xp += 1
            yp += 1
        elif x[xp][0] < y[yp][0] or \
                (x[xp][0] == y[yp][0] and x[xp][1] < y[yp][1]):
            xp += 1
        else:
            yp += 1

    return mask


@numba.njit
def cantor_pairing(x, y):
    """
    Computes the cantor pairing function for two integer, mapping two integers
    bijectively to a single one.

    The function is given as follows:

        p(x, y) = ((x + y) * (x + y + 1)) / 2 + y

    Parameters:
    -----------
    - x: int
        The first integer.
    - y: int
        The second integer.

    Returns:
    --------
    - int
        The result of the pairing function.
    """
    return ((x + y) * (x + y + 1)) // 2 + y



#
#  nn/functional/random.py
#  bxtorch
#
#  Created by Oliver Borchert on May 18, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch.distributions as dist

def generate_noise(shape, noise_type='normal'):
    """
    Generate noise which is usually fed to the generator of a GAN.

    Parameters:
    -----------
    - shape: torch.Size
        The shape of the noise to generate.
    - noise_type: str, default: 'normal'
        The type of the noise to generate. Must be one of the following:
        * normal: Draws from a standard normal distribution.
        * uniform: Draws uniformly from the range [-1, 1].

    Returns:
    --------
    - torch.FloatTensor
        The generated noise with the specified size.
    """
    if noise_type == 'normal':
        return dist.Normal(0, 1).sample(shape)
    elif noise_type == 'uniform':
        return dist.Uniform(-1, 1).sample(shape)
    raise ValueError(f'Invalid noise type {noise_type}.')


#
#  nn/data/composite.py
#  bxtorch
#
#  Created by Oliver Borchert on May 10, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

class ZipDataLoader:
    """
    Data loader which yields the elements of multiple data loaders.
    """

    pass



#
#  nn/callbacks/logging.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from .base import TrainingCallback, PredictionCallback
from bxtorch.utils.stdio import ProgressBar

class EpochProgressLogger(TrainingCallback):
    """
    Logs the training progress. It does only consider epochs (to plot the 
    progress of each batch within an epoch, use ``BatchProgressLogger``).
    """

    # MARK: Initialization
    def __init__(self, file=None):
        """
        Initializes a new progress printer for epochs.

        Parameters:
        -----------
        - file: str, default: None
            If given, the progress is not written to the command line, but to
            a file instead. Might be especially useful when multiple processes
            perform training simultaneously.
        """
        self.progress = None
        self.file = file

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        self.progress = ProgressBar(num_epochs, self.file)
        self.progress.start()

    def after_epoch(self, metrics):
        self.progress.step()

    def after_training(self):
        self.progress.finish()
        self.progress = None

    
class BatchProgressLogger(TrainingCallback):
    """
    Logs the training progress. It does only consider batches (to plot the 
    progress of the entire training, use ``EpochProgressPrinter``).
    """

    # MARK: Initialization
    def __init__(self, file=None):
        """
        Initializes a new progress printer for epochs.

        Parameters:
        -----------
        - file: str, default: None
            If given, the progress is not written to the command line, but to
            a file instead. Might be especially useful when multiple processes
            perform training simultaneously.
        """
        self.num_epochs = None
        self.progress = None
        self.file = file

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        self.num_epochs = num_epochs

    def before_epoch(self, current, num_iterations):
        print(f"Epoch {current+1}/{self.num_epochs}:")
        self.progress = ProgressBar(num_iterations, self.file)
        self.progress.start()

    def after_batch(self):
        self.progress.step()

    def after_epoch(self, metrics):
        self.progress.finish(metrics.to_dict())
        self.progress = None

    def after_training(self):
        self.num_epochs = None


class PredictionProgressLogger(PredictionCallback):
    """
    Logs the prediction progress.
    """

    # MARK: Initialization
    def __init__(self, file=None):
        """
        Initializes a new progress printer for predictions.

        Parameters:
        -----------
        - file: str, default: None
            If given, the progress is not written to the command line, but to
            a file instead. Might be especially useful when multiple processes
            perform training simultaneously.
        """
        self.progress = None
        self.file = file

    # MARK: Instance Methods
    def before_predictions(self, model, num_iterations):
        self.progress = ProgressBar(num_iterations, self.file)
        self.progress.start()

    def after_batch(self):
        self.progress.step()

    def after_predictions(self):
        self.progress.finish()
        self.progress = None



#
#  nn/training/parallel.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
import torch.nn as nn
import torch.multiprocessing as mp
import bxtorch.nn as xnn
from .wrappers import History
from bxtorch.utils.torch import gpu_device

class Hogwild:
    """
    The Hogwild class enables using the Hogwild training procedure when training
    a model. Only use this class if your epochs take reasonably long (e.g. a few
    minutes) and you require multiple epochs for training (e.g. > 10).
    Otherwise, speedups might be very marginal but computational costs greatly
    increased.

    Generally, it works as follows:

    1. For each epoch, a certain number of processes is started. These processes
        each train for a single epoch while updating the weights of the same
        model. Callbacks are only passed to the first process that has been
        started.
    2. Evaluation is only performed on the first of the processes. If the others
        have already finished while evaluating, they have to wait.
    3. If training continues, each process gets notified that training
        continues.
    """

    def __init__(self, trainer):
        """
        Initializes a new Hogwild context.

        Parameters:
        -----------
        - trainer: bxtorch.nn.BaseTrainer
            A trainer whose train function to use.
        """
        self.trainer = trainer

    def train(self, num_processes=None, gpu=False, **kwargs):
        """
        Trains the passed trainer's model using Hogwild.

        Parameters:
        -----------
        - gpu: bool or int or list of int, default: False
            The GPU(s) to use for training. If multiple GPUs are specified, they
            are distributed among the processes. Generally, you shouldn't have
            more processes than GPUs.
        - num_processes: int, default: None
            The number of processes to use for Hogwild. If set to None or 0, it
            defaults to the number of processors.
        - kwargs: keyword arguments
            Arguments as passed directly to a trainer's ``train`` method.
            Subclasses of the base trainer may, however, require some
            parameters specified differently.

        Returns:
        --------
        - bxtorch.nn.History
            The history of the training for the first process.
        """
        # 1) Prepare for Hogwild
        try:
            mp.set_start_method('spawn')
        except Exception as e:
            print(e)
        num_processes = num_processes or os.cpu_count()
        self.trainer.model.share_memory()

        # 2) Prepare training on multiple processes
        history_queue = mp.Queue()
        devices = [gpu_device(g) for g in gpu] if isinstance(gpu, list) else \
            [gpu_device(gpu)]
        device_idx = 0

        processes = []
        push_queues = []
        pull_queues = []
        for i in range(num_processes):
            h_queue = history_queue if i == 0 else None
            push_queue = mp.Queue()
            pull_queue = mp.Queue()

            kw = self.trainer.parallel_prepare(**kwargs)
            if i != 0:
                replace = {
                    'callbacks': [],
                    'val_data': None,
                    'val_data_init': None,
                    'eval_train': False
                }
                for k, v in replace.items():
                    kw[k] = v

            device = devices[device_idx]
            process = mp.Process(
                target=_hogwild_train,
                args=(h_queue, push_queue, pull_queue, self.trainer, device, kw)
            )
            process.start()

            device_idx = (device_idx + 1) % len(devices)
            processes.append(process)
            push_queues.append(push_queue)
            pull_queues.append(pull_queue)

        # 3) Perform training based on synchronization calls
        continue_training = True
        while continue_training:
            for q in push_queues:
                q.cancel_join_thread()
                q.put(True)

            for q in pull_queues:
                continue_training = continue_training and q.get()

        # 4) Finish training
        history = history_queue.get()
        history_queue.close()

        for q in push_queues:
            q.cancel_join_thread()
            q.put(False)
            q.close()

        # 5) Graceful shutdown
        for q in pull_queues:
            q.close()

        for p in processes:
            p.join()

        return history


def _hogwild_train(history_queue, push_queue, pull_queue, trainer, device, 
                   kwargs):
    kwargs = {
        **trainer.parallel_restore(**kwargs),
        'gpu': None
    }
    kwargs['callbacks'] = kwargs['callbacks'] + [
        xnn.SynchronizationCallback(push_queue, pull_queue)
    ]

    # Train
    trainer.device = device
    history = trainer.train(**kwargs)

    # Return history
    if history_queue is not None:
        history_queue.cancel_join_thread()
        history_queue.put(history)



#
#  nn/callbacks/early_stopping.py
#  bxtorch
#
#  Created by Oliver Borchert on May 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import copy
from .base import TrainingCallback, CallbackException

class EarlyStopping(TrainingCallback):
    """
    The early stopping callback watches a specified metric and interrupts
    training if the metric does not decrease for a specified number of epochs.
    """

    # MARK: Initialization
    def __init__(self, metric='val_loss', patience=5, restore_best=False,
                 minimize=True):
        """
        Initializes a new early stopping callback.

        Parameters:
        -----------
        - metric: str, default: 'val_loss'
            The metric to watch during training.
        - patience: int, default: 5
            The number of epochs that training still continues although the
            watched metric is greater than its smallest value during training.
        - restore_best: bool, default: False
            Whether the model's parameter should be set to the parameters which
            showed the best performance in terms of the watched metric.
        - minimize: bool, default: True
            Whether to minimize or maximize the given metric.
        """
        self.patience = patience
        self.epoch = None
        self.counter = None
        self.best_metric = None
        self.model = None
        self.state_dict = None
        self.restore_best = restore_best
        self.metric = metric
        self.minimize = minimize

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        self.model = model
        if self.restore_best:
            self.state_dict = copy.deepcopy(model.state_dict())
        self.epoch = 0
        self.counter = 0
        self.best_metric = float('inf') if self.minimize else -float('inf')

    def after_epoch(self, metrics):
        self.epoch += 1
        if self._is_metric_better(metrics):
            if self.restore_best:
                self.state_dict = copy.deepcopy(self.model.state_dict())
            self.counter = 0
            self.best_metric = metrics[self.metric]
        else:
            self.counter += 1
            if self.counter == self.patience:
                raise CallbackException(
                    f"Early stopping after epoch {self.epoch} (patience {self.patience})."
                )

    def after_training(self):
        if self.restore_best and self.counter > 0:
            self.model.load_state_dict(self.state_dict)

    # MARK: Private Methods
    def _is_metric_better(self, metrics):
        if self.minimize:
            return metrics[self.metric] < self.best_metric
        else:
            return metrics[self.metric] > self.best_metric



#
#  nn/training/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .parallel import Hogwild
from .supervised import SupervisedTrainer
from .trainer import BaseTrainer



#
#  graph/graph.py
#  bxtorch
#
#  Created by Oliver Borchert on June 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np

def symmetric_edge_list(edges):
    """
    Given a list of edges, this function computes a set which contains the
    edges "in both directions".

    Parameters:
    -----------
    - edges: numpy.ndarray [N, 2]
        The edges where edges are considered undirected.

    Returns:
    --------
    - numpy.ndarray [N * 2, 2]
        The edges where edges may be considered directed.
    """
    reverse_edges = np.column_stack([edges[:,1], edges[:,0]])
    return np.concatenate([edges, reverse_edges])



#
#  nn/callbacks/base.py
#  bxtorch
#
#  Created by Oliver Borchert on May 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from abc import ABC

class CallbackException(Exception):
    """
    Exception raised by callbacks to stop the training procedure.
    """
    pass


class TrainingCallback(ABC):
    """
    Abstract class to be subclassed by all training callbacks.
    These callbacks are passed to a trainer which calls the implemented methods
    at specific points during training.
    """

    # MARK: Instance Methods
    def before_training(self, model, num_epochs):
        """
        Method is called prior to the start of the training.
        This method must not raise exceptions.

        Parameters:
        -----------
        - model: torch.nn.Module
            The model which is trained.
        - num_epochs: int
            The maximum number of epochs performed during training.
        """
        pass

    def before_epoch(self, current, num_iterations):
        """
        Method is called at the begining of every epoch during training.
        This method may raise exceptions if training should be stopped. Note,
        however, that stopping training at this stage is an advanced scenario.

        Parameters:
        -----------
        - current: int
            The index of the epoch that is about to start.
        - num_iterations: int
            The expected number of iterations for the batch.
        """
        pass

    def after_batch(self):
        """
        Method is called at the end of a mini-batch. If the data is not
        partitioned into batches, this function is never called.
        The method may not raise exceptions.
        """
        pass

    def after_epoch(self, metrics):
        """
        Method is called at the end of every epoch during training.
        This method may raise exceptions if training should be stopped.

        Parameters:
        -----------
        - metrics: bxtorch.nn.training.wrappers.Evaluation
            Metrics obtained after training this epoch.
        """
        pass

    def after_training(self):
        """
        Method is called upon end of the training procedure.
        The method may not raise exceptions.
        """
        pass


class PredictionCallback(ABC):
    """
    Abstract class to be subclassed by all prediction callbacks.
    These callbacks are passed to a trainer which calls the implemented methods
    at specific points during inference.
    """

    # MARK: Instance Methods
    def before_predictions(self, model, num_iterations):
        """
        Called before prediction making starts.

        Parameters:
        -----------
        - model: torch.nn.Module
            The model which is used to make predictions.
        - num_iterations: int
            The number of iterations/batches performed for prediction.
        """
        pass

    def after_batch(self):
        """
        Called after prediction is done for one batch.
        """
        pass

    def after_predictions(self):
        """
        Called after all predictions have been made.
        """
        pass



#
#  job/arguments.py
#  bxtorch
#
#  Created by Oliver Borchert on May 15, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from abc import ABC, abstractmethod
import argparse

class _Argument(ABC):
    """
    Base class for describing a specific type of argument for a job.
    """

    def __init__(self, identifier, default=None, allow_list=False, help=None):
        """
        Initializes a new argument according to the given parameters:

        Parameters:
        -----------
        - identifier: str
            The name of the argument. Set on the command line by 
            ``--identifier`` and can be accessed on the job with the
            ``.identifier`` property.
        - default: type, default: None
            The default value for the argument. If not given, the argument is
            assumed to be required.
        - allow_list: bool, default: False
            Determines whether the argument can be set multiple times, resulting
            in a list of values.
        - help: str, default: None
            A description of the argument.
        """
        self.identifier = identifier
        self.default = default
        self.allow_list = allow_list
        self.help = help

    def register_on(self, parser):
        """
        Adds the argument to the specified parser.

        Parameters:
        -----------
        - parser: argparse.ArgumentParser
            The parser to add the argument to.
        """
        kwargs = {
            'required': self.default is None,
            'default': self.default,
            'help': self.help, 
            'action': 'append' if isinstance(self.default, list) \
                      else self._action
        }

        if kwargs['action'] not in ('store_true', 'store_false'):
            kwargs['type'] = self._from_string

        parser.add_argument(f'--{self.identifier}', **kwargs)

    @property
    def _action(self):
        return 'store'

    @abstractmethod
    def _from_string(self, value):
        pass


class Boolean(_Argument):
    """
    Argument of type boolean. 
    
    Note:
    -----
    You may not provide a value on the command line. Setting a boolean argument
    (i.e. flag) simply inverts the default value.
    """

    @property
    def _action(self):
        return 'store_false' if self.default else 'store_true'

    def _from_string(self, value):
        raise argparse.ArgumentTypeError(
            'Invalid option for boolean argument.'
        )


class Integer(_Argument):
    """
    Argument of type integer.
    """

    def _from_string(self, value):
        return int(value)


class String(_Argument):
    """
    Argument of type string.
    """

    def _from_string(self, value):
        return value


class Float(_Argument):
    """
    Argument of type float.
    """

    def _from_string(self, value):
        return float(value)



#
#  nn/callbacks/parallel.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
from .base import TrainingCallback, CallbackException

class SynchronizationCallback(TrainingCallback):
    """
    Synchronizer callback which can be used within a parallel environment to
    stop and resume training at specific points.
    """

    # MARK: Initialization
    def __init__(self, push_queue, pull_queue):
        """
        Initializes a new synchronizer.

        Parameters:
        -----------
        - push_queue: torch.multiprocessing.Queue
            The queue onto which the parallel environment pushes whether the
            current epoch should be trained. If ``False`` is passed, training
            is stopped.
        - pull_queue: torch.multiprocessing.Queue
            The queue onto which the callback pushes ``True`` to indicate that 
            the epoch is finished. Upon end of training, it pushes ``False``.
        """
        self.push_queue = push_queue
        self.pull_queue = pull_queue

    # MARK: Instance Methods
    def before_epoch(self, current, num_iterations):
        cont = self.push_queue.get()
        if not cont:
            raise CallbackException(
                f'Process {os.getpid()} is told to finish training.'
            )

    def after_epoch(self, metrics):
        self.pull_queue.cancel_join_thread()
        self.pull_queue.put(True)

    def after_training(self):
        self.pull_queue.cancel_join_thread()
        self.pull_queue.put(False)



#
#  nn/utils/predictor.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import os
import numpy as np
import torch
import torch.nn as nn
import torch.multiprocessing as mp
from torch.utils.data import DataLoader, Dataset
from bxtorch.utils.torch import gpu_device, to_device
import bxtorch.multiprocessing as xmp

class Predictor:
    """
    Given a model, the predictor can be used to easily compute predictions for
    a model. It can be seen as binding between raw data and model by using
    a dataset and enabling parallelism.
    """

    # MARK: Initialization
    def __init__(self, model, dataset_class=None, **dataset_kwargs):
        """
        Initializes a new predictor with the given parameters.

        Parameters:
        -----------
        - model: torch.nn.Module
            The model to make predictions with.
        - dataset_class: type like torch.utils.data.Dataset, default: None
            The dataset class to use to make predictions. It must accept as
            first parameter a list of raw elements for prediction. In case no
            class is defined, the ``predict`` function expects to receive a
            dataset instead of raw elements.
        - dataset_kwargs: keyword arguments
            Arguments passed to the dataset upon initialization.
        """
        self.model = model
        self.dataset_class = dataset_class
        self.dataset_kwargs = dataset_kwargs

    # MARK: Instance Methods
    def predict(self, samples, callbacks=[], gpu=False, **kwargs):
        """
        Computes predictions for the given samples.

        Parameters:
        -----------
        - samples: list of object
            If a list of objects is given, the samples to make predictions for
            are fed directly to the dataset initializer as first parameters. If
            a dataset is given, the dataset will be used to make predictions.
        - callbacks: list of bxtorch.nn.PredictionCallback
            Callbacks which are called as prediction progresses.
        - gpu: bool or int or list of int, default: False
            Whether to use a (specific) GPU or multiple GPUs. If multiple GPUs
            are used, one process per GPU is started to minimize
            synchronization. Make sure that using multiple GPUs makes up for
            this overhead. If ``False`` is specified, all cores of the 
            computer are used to make predictions in parallel.
        - kwargs: keyword arguments
            Additional arguments fed directly to the data loader. Includes e.g.
            ``batch_size``.

        Returns:
        --------
        - numpy.ndarray
            The predictions made by the model.
        """
        if isinstance(samples, Dataset):
            dataset = samples
        else:
            dataset = self.dataset_class(samples, **self.dataset_kwargs)

        if hasattr(dataset, 'loader'):
            loader = dataset.loader(**kwargs)
        else:
            if isinstance(gpu, list) or gpu == False:
                kwargs['pin_memory'] = True
            kwargs['shuffle'] = False
            loader = DataLoader(dataset, **kwargs)

        num_iterations = len(loader)

        self._exec_callbacks(
            callbacks, 'before_predictions', self.model, num_iterations
        )

        self.model.eval()

        if isinstance(gpu, list) or gpu == False: # parallel computation
            num_gpus = len(gpu) if isinstance(gpu, list) else os.cpu_count()

            self.model.share_memory()

            callback = lambda:self._exec_callbacks(callbacks, 'after_batch')
            vectorizer = xmp.Vectorizer(
                _worker_func, _worker_init, callback_func=callback,
                num_workers=num_gpus, gpu=gpu, model=self.model
            )

            predictions = vectorizer.process(
                loader, self.model, self._process_prediction
            )
        else: # sequential computation
            device = gpu_device(gpu)
            self.model.to(device)

            predictions = []

            for x in loader:
                out = _worker_func(
                    x, self.model, self._process_prediction, device
                )
                predictions.append(out)
                self._exec_callbacks(
                    callbacks, 'after_batch'
                )

        self._exec_callbacks(
            callbacks, 'after_predictions'
        )

        return self._collate_predictions(predictions)

    # MARK: Private Methods
    def _process_prediction(self, x, out):
        return out

    def _collate_predictions(self, out):
        return np.concatenate([p.numpy() for p in out])

    def _exec_callbacks(self, callbacks, func, *args):
        for callback in callbacks:
            getattr(callback, func)(*args)


def _worker_func(item, model, process, device):
    x = to_device(device, item)
    with torch.no_grad():
        if isinstance(x, (list, tuple)):
            out = model(*x)
        else:
            out = model(x)
    return to_device('cpu', process(x, out))


def _worker_init(rank, gpu, model):
    if isinstance(gpu, list):
        device = gpu_device(gpu[rank])
    else:
        device = gpu_device(gpu)
    model.to(device)
    return device



#
#  nn/data/infinite_dataset.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from abc import ABC
import torch.utils.data as torch_data
from .infinite_loader import InfiniteDataLoader

class InfiniteDataset(ABC):
    """
    The infinite dataset works similarly to PyTorch's Dataset. However, instead
    of having a fixed size, it yields infinitely many items via the ``next()``
    method.
    """

    # MARK: Initialization
    def __init__(self, generator):
        """
        Initializes a new infinite dataset.

        Parameters:
        -----------
        - generator: generator
            The generator yielding new items.
        """
        self.generator = generator

    # MARK: Instance Methods
    def loader(self, **kwargs):
        """
        Returns a data loader for this dataset.

        Parameters:
        -----------
        - kwargs: keyword arguments
            Paramaters passed directly to the DataLoader.

        Returns:
        --------
        - bxtorch.nn.InfiniteDataLoader
            The data loader with the specified attributes.
        """
        if hasattr(self, 'collate_fn'):
            kwargs['collate_fn'] = self.collate_fn
        return InfiniteDataLoader(self, **kwargs)

    # MARK: Private Methods
    def _transform(self, item):
        """
        Transforms the item retrieved from a generator to an object picked up
        by the data loader.
        The default implementation just returns the item as is.

        Parameters:
        -----------
        - item: generator item
            The item provided by the dataset's generator.

        Returns:
        --------
        - object
            The object returned for the data loader.
        """
        return item

    # MARK: Special Methods
    def __iter__(self):
        return self

    def __next__(self):
        return self._transform(next(self.generator))



#
#  nn/training/trainer.py
#  bxtorch
#
#  Created by Oliver Borchert on May 09, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import copy
from abc import ABC, abstractmethod
import time
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from bxtorch.nn.callbacks import CallbackException, TrainingCallback, \
    PredictionCallback
from .wrappers import History, Evaluation
from bxtorch.utils.torch import gpu_device, to_device
from bxtorch.nn.data import InfiniteDataLoader, InfiniteDataset
from bxtorch.nn.callbacks.base import TrainingCallback, PredictionCallback

class BaseTrainer(TrainingCallback, PredictionCallback, ABC):
    """
    A base class for training and evaluating models. Generally, this class
    should be seen as "binding" between a model and some data. Models should
    always be wrapped in a trainer, both when being trained, and when performing
    inference (i.e. evaluation). The trainer ensures that the model's 
    environment is correct and prevents plenty of pitfalls.

    A concrete implementation of this class is tailored to a specific type of 
    data (e.g. independent, identically distributed data samples) and/or model
    types (e.g. GANs).
    """

    # MARK: Initialization
    def __init__(self, model):
        """
        Initializes a new trainer for a specified model.

        Parameters:
        -----------
        - model: torch.nn.Module
            The model to train or evaluate.
        """
        self.model = model
        self.device = None
        self._cache = {}
        self._iteration = None

    # MARK: Instance Methods
    def optimizer(self, optimizer='adam', **kwargs):
        """
        Convenience function to create an optimizer for the model from a string.

        Parameters:
        -----------
        - optimizer: str, default: 'adam'
            The optimizer to use. Acceptable values are all optimizers from
            torch.optim (upper and lowercase does not matter).
        - kwargs: keyword arguments
            Arguments passed to the optimizer during initalization.
        """
        valid_optimizers = [
            'ASGD', 'Adadelta', 'Adagrad', 'Adam', 'Adamax', 'LBFGS',
            'RMSprop', 'Rprop', 'SGD', 'SparseAdam'
        ]
        for opt in valid_optimizers:
            if opt.lower() == optimizer.lower():
                return getattr(optim, opt)(self.model.parameters(), **kwargs)
        raise ValueError(f'Invalid optimizer {optimizer}.')

    def train(self, train_data, val_data=None, epochs=20, val_iterations=None,
              eval_every=1, eval_train=False, callbacks=[], metrics={},
              gpu=False, **kwargs):
        """
        Method for training the model with the supplied parameters.

        Parameters:
        -----------
        - train_data: torch.DataLoader or bxtorch.nn.InfiniteDataLoader
            A data loader to obtain training data from. The samples yielded by
            the data loader depend on a specific trainer implementation.
        - val_data: torch.DataLoader or bxtorch.nn.InfiniteDataLoader, 
                default: None
            A data loader to use for validation. If the loader is an infinite
            data loader, ``val_iterations`` must also be given. If not supplied,
            no validation will be performed.
        - epochs: int, default: 20
            The number of epochs to train for. If the given data is an infinite
            data loader, this value defines the number of iterations (and 
            should most probably be increased).
        - val_iterations: int, default: None
            The number of iterations to perform for validation. Must be given
            only if ``val_data`` is an infinite data loader. Otherwise, the
            parameter is ignored.
        - eval_every: int, default: 1
            How many iterations should pass until validation is called again.
            This should only be set if ``train_data`` is an infinite loader,
            otherwise it will be ignored. Note that everything works as if
            ``epochs // eval_every`` epochs is trained while ``eval_every`` can
            be considered the number of mini-batches.
        - eval_train: bool, default: False
            Whether to compute metrics (apart from the loss) for both the
            validation and the train data. If this flag is set to False,
            metrics will only be computed for the validation data.
        - callbacks: list of bxtorch.nn.TrainingCallback or 
                bxtorch.nn.PredictionCallback, default: []
            The callbacks to use for training and inference. The training
            callbacks and prediction callbacks will be filtered automatically.
        - metrics: dict of str -> func, default: {}
            Metrics to compute during evaluation for the validation data (and
            potentially for the training data). The keys for the metrics define
            their name.
        - gpu: bool or int or list of int, default: False
            Governs, whether training and evaluation should be performed on a
            GPU. If set to True, the GPU with the most amount of free memory is
            selected (if there are multiple GPUs). If set to an integer, the GPU
            with the specified index is used. If set to a list of integers, the
            specified GPUs are used to train and evaluate the model und multiple
            GPUs simultaneously. In this case, the batch sizes of the data
            loaders should be adjusted accordingly. If set to None, the model
            won't be moved. Only do this if you know what you are doing.
        - kwargs: keyword arguments
            Additional keyword arguments dependent on the specific subclass.

        Returns:
        --------
        - bxtorch.nn.training.wrappers.History
            A history object summarizing stats from the training. It contains
            as properties the development of the loss as ``train_loss`` (and
            potentially ``val_loss``, if ``val_data`` is supplied and the 
            keyword arguments include a function called ``loss``). If
            additional metrics are supplied, there will be a property 
            ``val_<metric>`` for each metric, and potentially ``train_<metric>``
            if ``eval_train`` is set to True.
        """
        # 1) Setup
        exception = None

        if isinstance(train_data, InfiniteDataLoader):
            epochs = epochs // eval_every

        # 1.1) Callbacks
        train_callbacks = [
            c for c in callbacks if isinstance(c, TrainingCallback)
        ]
        prediction_callbacks = [
            c for c in callbacks if isinstance(c, PredictionCallback)
        ]

        self._exec_callbacks(
            train_callbacks, 'before_training', self.model, epochs
        )

        # 1.2) Metrics
        metric_history = []
        if 'loss' in kwargs:
            val_metrics = {**metrics, **{'loss': kwargs['loss']}}
        else:
            val_metrics = metrics

        # 1.3) Data loading
        train_epoch_is_iteration = not isinstance(train_data, DataLoader)
        if train_epoch_is_iteration:
            train_iterator = iter(train_data)

        # 1.4) GPU support
        if gpu is not None:
            self._setup_device(gpu)
            self.model.to(self.device)
        
        # 2) Train for number of epochs
        for current_epoch in range(epochs):
            # 2.1) Prepare
            if train_epoch_is_iteration:
                batch_iterations = eval_every
            else:
                batch_iterations = len(train_data)
                
            try:
                self._exec_callbacks(
                    train_callbacks, 'before_epoch', current_epoch, 
                    batch_iterations
                )
            except CallbackException as e:
                exception = e
                break

            if not train_epoch_is_iteration:
                train_iterator = iter(train_data)

            # 2.2) Train
            self.model.train()

            train_batch_weights = []
            train_losses = []

            if train_epoch_is_iteration:
                # 2.2.1) Infinite iterator
                for _ in range(eval_every):
                    item = next(train_iterator)
                    item = to_device(self.device, item)
                    loss = self._train_batch(item, **kwargs)
                    train_losses.append(loss)
                    self._exec_callbacks(
                        train_callbacks, 'after_batch'
                    )
            else:
                # 2.2.2) Dataset
                for item in train_iterator:
                    item = to_device(self.device, item)
                    loss = self._train_batch(item, **kwargs)
                    train_batch_weights.append(len(item))
                    train_losses.append(loss)
                    self._exec_callbacks(
                        train_callbacks, 'after_batch'
                    )

            # 2.3) Validate
            batch_metrics = Evaluation(
                self._collate_train_losses(train_losses),
                train_batch_weights
            )

            if val_data is not None:
                eval_val = self.evaluate(
                    val_data, iterations=val_iterations, metrics=val_metrics,
                    callbacks=prediction_callbacks, gpu=None
                ).with_prefix('val_')
                batch_metrics = Evaluation.merge(batch_metrics, eval_val)

            if eval_train:
                eval_train = self.evaluate(
                    train_data, iterations=val_iterations, metrics=metrics,
                    callbacks=prediction_callbacks, gpu=None
                ).with_prefix('train_')
                batch_metrics = Evaluation.merge(batch_metrics, eval_train)

            batch_metrics = Evaluation.merge(
                batch_metrics, Evaluation({'_timestamp': time.time()})
            )

            metric_history.append(batch_metrics)

            # 2.4) Finish epoch
            try:
                self._exec_callbacks(
                    train_callbacks, 'after_epoch', batch_metrics
                )
            except CallbackException as e:
                exception = e
                break

        # 3) Finish training

        # 3.1) If GPU used
        if gpu is not None:
            self.model.to('cpu')
            self.device = None

        # 3.2) Finish callbacks
        self._exec_callbacks(
            train_callbacks, 'after_training'
        )
        if exception is not None:
            print(exception)

        return History(metric_history)

    def evaluate(self, data, iterations=None, metrics={}, callbacks=[],
                 gpu=False):
        """
        Evaluates the model on the given data and computes the supplied metrics.

        Parameters:
        -----------
        - data: torch.DataLoader or bxtorch.nn.InfiniteDataLoader
            A data loader to obtain evaluation samples from. The expected 
            samples depend on a specific trainer subclass.
        - iterations: int, default: None
            The number of samples used for evaluating if the given data is an
            infinite data loader.
        - metrics: dict of str -> func, default: {}
            The metrics to evaluate the model for. The keys define the names of
            the metrics when retrieving the evaluated result from the return
            parameter.
        - callbacks: list of bxtorch.nn.PredictionCallback, default: []
            Callbacks to use while computing predictions. Usually, they are 
            used for logging.
        - gpu: bool or int or list of int, default: False
            Governs, whether training and evaluation should be performed on a
            GPU. If set to True, the GPU with the most amount of free memory is
            selected (if there are multiple GPUs). If set to an integer, the GPU
            with the specified index is used. If set to a list of integers, the
            specified GPUs are used to train and evaluate the model und multiple
            GPUs simultaneously. In this case, the batch sizes of the data
            loaders should be adjusted accordingly. If set to None, the model
            won't be moved. Only do this if you know what you are doing.

        Returns:
        --------
        - bxtorch.nn.training.wrappers.Evaluation
            An evaluation object, yielding as properties the metrics with their
            specified names.
        """
        num_predictions = iterations or len(data)
        self._exec_callbacks(
            callbacks, 'before_predictions', self.model, num_predictions
        )

        if gpu is not None:
            self._setup_device(gpu)
            self.model.to(self.device)

        self.model.eval()

        predictions = []
        targets = []

        iterator = iter(data)
        for _ in range(num_predictions):
            item = next(iterator)
            item = to_device(self.device, item)

            with torch.no_grad():
                prediction, target = self._predict_batch(item)

            predictions.append(to_device('cpu', prediction))
            targets.append(to_device('cpu', target))

            self._exec_callbacks(
                callbacks, 'after_batch'
            )

        self._exec_callbacks(
            callbacks, 'after_predictions'
        )
        
        predictions = self._collate_predictions(predictions)
        targets = self._collate_targets(targets)

        result = {
            k: f(predictions, targets).item()
            for k, f in metrics.items()
        }

        if gpu is not None:
            self.model.to('cpu')
            self.device = None

        return Evaluation(result)

    def parallel_prepare(self, **kwargs):
        """
        The parallel prepare function prepares the trainer for being passed to
        another thread. It ensures that the trainer's model is moved into
        shared memory.
        
        Parameters:
        -----------
        - kwargs: keyword arguments
            The parameters passed to the ``train`` function. All parameters
            which are not thread safe (such as data loaders or optimizers)
            must be passed as ``<parameter>_init`` and will then be initialized
            on the background thread.

        Returns:
        --------
        - dict
            The parameters which can safely be passed to a background thread.
        """
        self.model.share_memory()
        return kwargs

    def parallel_restore(self, **kwargs):
        """
        Based on the return values of the ``parallel_prepare`` function, this
        function should use the passed parameters to construct the actual
        parameters passed to the ``train`` function.

        Parameters:
        -----------
        - kwargs: keyword arguments
            The parameters retrieved from the ``parallel_prepare`` function.
            All parameters which are passed as ``<parameters>_init`` will be
            converted into the parameter ``<parameter>``. They are expected to
            be functions which take the model as first parameter.

        Returns:
        --------
        - dict
            The parameters to pass to the train function.
        """
        result = {}

        for k, v in kwargs:
            if k.endswith('_init'):
                result[k[:-5]] = v(self.model)
            else:
                result[k] = v

        return result

    # MARK: Private Methods
    @abstractmethod
    def _train_batch(self, data, **kwargs):
        """
        Runs a single step in training. If the training data represents an 
        infinite dataset, this equals a single iteration, otherwise a 
        mini-batch.

        Parameters:
        -----------
        - data: object
            The data for the current iteration/mini-batch.

        Returns:
        --------
        - object
            The loss computed for the batch. If the returned value is not float,
            overwrite ``_collate_train_losses``.
        """
        pass

    @abstractmethod
    def _predict_batch(self, data, **kwargs):
        """
        Runs a single step for inference. The data is either a mini-batch or a 
        single iteration, depending on the data used for evaluation.

        Parameters:
        -----------
        - data: object
            The data for the current iteration/mini-batch.
        - kwargs: keyword arguments
            Additional arguments dependent on the subclass and passed directly
            from the evaluation method.

        Returns:
        --------
        - torch.Tensor
            The output from the model.
        - object
            The target, i.e. correct output. If this is not a ``torch.Tensor``, 
            the ``_collate_targets`` should be overriden.
        """
        pass

    def _collate_train_losses(self, losses):
        """
        Combines the losses obtained from the ``_train_batch`` function.
        The default implementation assumes that simple floats are returned.

        Parameters:
        -----------
        - losses: list of object
            The losses returned from ``_train_batch``.
        
        Returns:
        --------
        - dict of str -> (float or list of float)
            The loss names mapped to their values.
        """
        return {'train_loss': losses}

    def _collate_predictions(self, predictions):
        """
        Combines the predictions obtained from the ``_predict_batch`` function.
        The default implementation assumes that predictions are tensors and can 
        simply be conatenated.

        Parameters:
        -----------
        - list of objects
            The predictions.

        Returns:
        --------
        - object
            An object to be used as predicted value for some metric.
        """
        return torch.cat(predictions)

    def _collate_targets(self, targets):
        """
        Combines the targets in a way that it can be passed to some metric.
        The default implementation assumes that targets are tensors and simply
        concatenates them.

        Parameters:
        -----------
        - list of object
            The targets.

        Returns:
        --------
        - object
            An object to be used as true target with some metric.
        """
        return torch.cat(targets)

    def _setup_device(self, gpu):
        if isinstance(gpu, list):
            self.model = nn.DataParallel(self.model, device_ids=gpu)
            self.device = gpu_device(gpu[0])
        else:
            self.device = gpu_device(gpu)

    def _exec_callbacks(self, callbacks, func, *args):
        for callback in ([self] + callbacks):
            getattr(callback, func)(*args)



#
#  nn/data/dataset.py
#  bxtorch
#
#  Created by Oliver Borchert on May 21, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import torch.utils.data as torch_data

class Dataset(torch_data.Dataset):
    """
    Subclass of a simple PyTorch dataset which includes an additional
    ``loader`` function such that subclasses can directly initialize data
    loaders. If the dataset defines a function named ``collate_fn``, it is
    used instead of the default collate function.
    """

    def loader(self, **kwargs):
        """
        Returns a data loader for this dataset.

        Parameters:
        -----------
        - kwargs: keyword arguments
            Paramaters passed directly to the DataLoader.

        Returns:
        --------
        - torch.utils.data.DataLoader
            The data loader with the specified attributes.
        """
        if hasattr(self, 'collate_fn'):
            kwargs['collate_fn'] = self.collate_fn
        return torch_data.DataLoader(self, **kwargs)



#
#  setup.py
#  bxtorch
#
#  Created by Oliver Borchert on May 23, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='BxTorch',
    version='0.2.25',

    author='Oliver Borchert',
    author_email='borchero@in.tum.de',

    description='High-level abstractions for PyTorch.',
    long_description=long_description,
    long_description_content_type = 'text/markdown',

    url='https://github.com/borchero/bxtorch',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ],
    python_requires='>=3.7',
    install_requires=[
        'torch>=1.1.0,<2.0.0',
        'numpy>=1.16.3,<2.0.0',
        'scipy>=1.3.0,<2.0.0',
        'numba>=0.43.1,<1.0.0',
        'scikit-learn>=0.20.3,<0.21.0'
    ],

    license='License :: OSI Approved :: MIT License',
    zip_safe=False
)



#
#  nn/training/wrappers.py
#  bxtorch
#
#  Created by Oliver Borchert on May 20, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    pass

class History:
    """
    This class summarizes metrics obtained during training of a model.
    """

    # MARK: Static Methods
    @staticmethod
    def cat(hist1, hist2):
        """
        Concatenates the two history objects.

        Parameters:
        -----------
        - hist1: bxtorch.nn.History
            The first history.
        - hist2: bxtorch.nn.History
            The second history.

        Returns:
        --------
        - bxtorch.nn.History
            The concatenated history object.
        """
        return History(hist1._metrics + hist2._metrics)
    
    # MARK: Initialization
    def __init__(self, metrics):
        """
        Initializes a new history object.

        Parameters:
        -----------
        - metrics: list of dict
            The metrics from each training step.
        """
        self._metrics = metrics

    # MARK: Instance Methods
    def plot(self, only=None, save_to=None):
        """
        Plots the history via matplotlib. If running in a Jupyter notebook, this
        call should be followed by a call to ``plt.show()``. For all metrics
        ``<metric>``, the function combines ``val_<metric>`` and
        ``train_<metric>`` (if available) into plots, yet uses different
        subplots for different metrics.

        Parameters:
        -----------
        - only: list of str, default: None
            If this value is defined, the history is only visualized for the
            given metrics. 
        - save_to: str, default: None
            If given, the generated plot is saved to the given file.
        """
        x = np.arange(len(self))
        raise NotImplementedError

    # MARK: Special Methods
    def __getstate__(self):
        return self._metrics

    def __setstate__(self, state):
        self._metrics = state

    def __len__(self):
        return len(self._metrics)

    def __getattr__(self, name):
        if len(self._metrics) == 0:
            raise AttributeError(
                f'Length of history is 0. Metric {name} cannot be accessed.'
            )
        if name in self._metrics[0]:
            return [m[name] for m in self._metrics]
        else:
            raise AttributeError(f'Metric {name} does not exist.')
            

class Evaluation:
    """
    This class summarizes metrics obtained when evaluating a model.
    """

    # MARK: Static Methods
    @staticmethod
    def merge(eval1, eval2, *eval_other):
        """
        Merges multiple evaluations into one.

        Parameters:
        -----------
        - eval1: torch.nn.training.wrappers.Evaluation
            The first evaluation.
        - eval2: torch.nn.training.wrappers.Evaluation
            The second evaluation.
        - eval_other: varargs of torch.nn.training.wrappers.Evaluation
            Additional evaluations.

        Returns:
        --------
        - torch.nn.training.wrappers.Evaluation
            The merged evaluation.
        """
        result = Evaluation({})
        metrics = {}
        for e in [eval1, eval2] + list(eval_other):
            for k, v in e._metrics.items():
                metrics[k] = v
        result._metrics = metrics
        return result

    # MARK: Initialization
    def __init__(self, metrics, weights=None):
        """
        Initializes a new evaluation wrapper from the given metrics.

        Parameters:
        -----------
        - metrics: dict of str -> list of float
            The metrics computed. The keys define the metric names, the values
            provide the metrics computed over all batches.
        - weights: list of int, default: None
            The length of the batches. Should be supplied if the batches for
            which the metrics have been computed have differing sizes. If so,
            this list must contain the sizes of all batches.
        """
        result = {}
        for metric, values in metrics.items():
            if weights is None or len(weights) == 0:
                average = np.mean(values)
            else:
                weights = np.array(weights)
                values = np.array(values)
                average = weights.dot(values) / np.sum(weights)
            result[metric] = average
        self._metrics = result

    # MARK: Instance Methods
    def to_dict(self):
        """
        Returns the evaluation object as plain dictionary.

        Returns:
        --------
        - dict
            The dictionary.
        """
        return {k: v for k,v in self._metrics.items() if not k.startswith('_')}

    def with_prefix(self, prefix):
        """
        Initializes a new evaluation wrapper where all metrics are prefixed
        with the given prefix. Useful when merging evaluations.

        Parameters:
        -----------
        - prefix: str
            The prefix to use.

        Returns:
        --------
        - torch.nn.training.wrappers.Evaluation
            The resulting evaluation.
        """
        result = Evaluation({})
        result._metrics = {f'{prefix}{k}': v for k, v in self._metrics.items()}
        return result

    # MARK: Special Methods
    def __getstate__(self):
        return self._metrics

    def __setstate__(self, state):
        self._metrics = state

    def __contains__(self, item):
        return item in self._metrics

    def __getitem__(self, item):
        return self._metrics[item]

    def __getattr__(self, name):
        try:
            return self._metrics[name]
        except KeyError:
            class_name = self.__class__.__name__
            raise AttributeError(
                f'{class_name} does not provide metric {name}.'
            )



#
#  multiprocessing/vectorize.py
#  bxtorch
#
#  Created by Oliver Borchert on June 02, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

import os
import numpy as np
import torch
import torch.multiprocessing as mp

class Vectorizer:
    """
    The Vectorizer class ought to be used in cases where a result tensor
    of size N is filled with values computed in some complex way. The
    computation of these N computations can then be parallelized over
    multiple processes.
    """

    # MARK: Initialization
    def __init__(self, worker_func, worker_init=None, callback_func=None,
                 num_workers=-1, **kwargs):
        """
        Initializes a new vectorizer.

        Parameters:
        -----------
        - worker_func: callable
            The function which receives as input an item of the input to process
            and outputs a value which ought to be returned.
        - worker_init: callable, default: None
            The function receives as input the rank of the worker (i.e. every
            time this function is called, it is called with a different integer
            as parameter). Its return values are passed as *last* parameters
            to ``worker_func`` upon every invocation.
        - callback_func: callable, default: None
            A function to call after every item has been processed. Must not
            need to be a free function as it is called on the main thread.
        - num_workers: int, default: -1
            The number of processes to use. If set to -1, it defaults to the
            number of available cores. If set to 0, everything is executed on
            the main thread.
        - kwargs: keyword arguments
            Additional arguments passed to the worker initialization function.
        """
        self.num_workers = os.cpu_count() if num_workers == -1 else num_workers
        self.worker_func = worker_func
        self.worker_init = worker_init
        self.callback_func = callback_func
        self.init_kwargs = kwargs

    # MARK: Instance Methods
    def process(self, items, *args):
        """
        Uses the vectorizer's worker function in order to process all items
        in parallel.

        Parameters:
        -----------
        - items: list of object or iterable of object
            The items to be processed by the workers. If given as iterable only
            (i.e. it does not support index access), the performance might
            suffer slightly due to an increased number of synchronizations.
        - args: variadic arguments
            Additional arguments passed directly to the worker function.

        Returns:
        --------
        - list of object
            The output generated by the worker function for each of the input
            items.
        """
        if self.num_workers == 0: # execute sequentially
            result = []
            init = _init_if_needed(self.worker_init, 0, **self.init_kwargs)
            all_args = _combine_args(args, init)
            for item in items:
                result.append(self.worker_func(item, *all_args))
            return result
        else:
            process_batches = hasattr(items, '__getitem__')
            
            if process_batches:
                return self._process_batches(items, *args)
            else:
                return self._process_consumers(items, *args)

    # MARK: Private Methods
    def _process_batches(self, items, *args):
        num_items = len(items)

        splits = np.array_split(np.arange(num_items), self.num_workers)
        splits = [0] + [a[-1] + 1 for a in splits]

        result = []

        processes = []
        queues = []
        done = mp.Event()
        if self.callback_func is not None:
            tick_queue = mp.Queue()
        else:
            tick_queue = None

        for i in range(self.num_workers):
            queue = mp.Queue()
            process = mp.Process(
                target=_batch_worker,
                args=(
                    queue, done, tick_queue, i, 
                    self.worker_func, self.worker_init,
                    self.init_kwargs, items[splits[i]:splits[i+1]],
                    *args
                )
            )
            process.start()
            processes.append(process)
            queues.append(queue)

        if self.callback_func is not None:
            for _ in range(num_items):
                tick_queue.get()
                self.callback_func()

        for i, q in enumerate(queues):
            result.extend(q.get())
            q.close()

        done.set()

        for p in processes:
            p.join()

        return result

    def _process_consumers(self, items, *args):
        result = []

        processes = []
        push_queue = mp.Queue()
        pull_queue = mp.Queue()

        for i in range(self.num_workers):
            process = mp.Process(
                target=_consumer_worker,
                args=(push_queue, pull_queue, i,
                      self.worker_func, self.worker_init, self.init_kwargs, 
                      *args)
            )
            process.start()
            processes.append(process)

        iterator = iter(items)
        index = 0
        expect = 0
        try:
            for _ in range(self.num_workers):
                item = next(iterator)
                expect += 1
                push_queue.cancel_join_thread()
                push_queue.put((index, item))
                index += 1

            while True:
                result.append(pull_queue.get())
                self.callback_func()
                expect -= 1
                item = next(iterator)
                expect += 1
                push_queue.cancel_join_thread()
                push_queue.put((index, item))
                index += 1

        except StopIteration:
            for _ in range(expect):
                result.append(pull_queue.get())
                self.callback_func()

            pull_queue.close()

            for _ in range(len(processes)):
                push_queue.cancel_join_thread()
                push_queue.put(None)

            push_queue.close()

            for p in processes:
                p.join()

        return [r[1] for r in sorted(result, key=lambda r:r[0])]


def _batch_worker(push_queue, done, tick_queue, rank,
                  worker_func, worker_init, init_kwargs, items, *args):

    init = _init_if_needed(worker_init, rank, **init_kwargs)
    all_args = _combine_args(args, init)

    result = []

    for i, item in enumerate(items):
        result.append(worker_func(item, *all_args))
        if tick_queue is not None:
            tick_queue.cancel_join_thread()
            tick_queue.put(None)

    push_queue.cancel_join_thread()
    push_queue.put(result)
    done.wait()


def _consumer_worker(pull_queue, push_queue, rank, 
                     worker_func, worker_init, init_kwargs, *args):

    init = _init_if_needed(worker_init, rank, **init_kwargs)
    all_args = _combine_args(args, init)

    while True:
        item = pull_queue.get()
        if item is None:
            break

        idx, item = item
        result = worker_func(item, *all_args)

        push_queue.cancel_join_thread()
        push_queue.put((idx, result))
    

def _init_if_needed(init, rank, **kwargs):
    if init is None:
        return None
    else:
        return init(rank, **kwargs)


def _combine_args(a, b):
    if b is None:
        return a
    elif isinstance(b, tuple):
        return a + b
    else:
        return a + (b,)



#
#  nn/data/__init__.py
#  bxtorch
#
#  Created by Oliver Borchert on May 19, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#

from .dataset import Dataset
from .infinite_dataset import InfiniteDataset
from .infinite_loader import InfiniteDataLoader
from .noise import NoiseDataset



#
#  job/job.py
#  bxtorch
#
#  Created by Oliver Borchert on May 15, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

from abc import ABC, abstractmethod
import argparse
import json

class Job(ABC):
    """
    An abstract class which should be subclassed to provide an implementation
    for a specific job to be run.
    """

    # MARK: Initialization
    def __init__(self, description, arguments):
        """
        Initializes a new job with the given arguments.

        Parameters:
        -----------
        - description: str
            The description of the job on the command line.
        - arguments: list of bxtorch.job._Argument
            The arguments for the job.
        """
        parser = argparse.ArgumentParser(
            description=description
        )
        identifiers = []
        for arg in arguments:
            arg.register_on(parser)
            identifiers.append(arg.identifier)
        self._args = parser.parse_args()
        self._identifiers = identifiers

    # MARK: Instance Methods
    @abstractmethod
    def run(self):
        pass

    def save_config(self, file):
        """
        Saves the configuration parameters passed to this job into the given
        file.

        Parameters:
        -----------
        - file: str
            The file to save the parameters to as JSON.
        """
        config = {i: getattr(self, i) for i in self._identifiers}
        with open(file, 'w+') as f:
            json.dump(config, f)

    # MARK: Special Methods
    def __getattr__(self, name):
        return getattr(self._args, name)



#
#  utils/stdio.py
#  bxtorch
#
#  Created by Oliver Borchert on May 11, 2019.
#  Copyright (c) 2019 Oliver Borchert. All rights reserved.
#  

class cached_property:
    """
    Python property which is computed exactly once and a cached value is
    returned upon every subsequent call.
    """

    # MARK: Initialization
    def __init__(self, func):
        self.func = func
        self._cache = None

    # MARK: Special Methods
    def __get__(self, obj, objtype):
        if self._cache is not None:
            return self._cache
        self._cache = self.func(obj)
        return self._cache

    def __repr__(self):
        return self.func.__doc__


def flatten(l):
    """
    Flattens the specified list.

    Parameters:
    -----------
    - l: list
        A two-dimensional list.

    Returns:
    --------
    - list
        The flattened list.
    """
    return [e for s in l for e in s]
